import json
import argparse
from datetime import datetime
from functools import partial
import os

import numpy as np
import pandas as pd
from sklearn.naive_bayes import GaussianNB
from sklearn.svm import SVC, SVR
import xgboost as xgb
from scipy.stats import spearmanr, kendalltau, weightedtau
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.neighbors import KNeighborsRegressor, KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from syntherela.metadata import Metadata
from syntherela.data import load_tables, remove_sdv_columns
from syntherela.metrics.utility import MachineLearningEfficacyMetric
from syntherela.utils import NpEncoder


## DATA LOADING
def load_rossmann(method):
    dataset_name = 'rossmann_subsampled' 
    run = "1"
    metadata = Metadata().load_from_json(f'{PROJECT_PATH}/data/original/{dataset_name}/metadata.json')

    tables = load_tables(f'{PROJECT_PATH}/data/original/{dataset_name}/', metadata)
    tables_synthetic = load_tables(f'{PROJECT_PATH}/data/synthetic/{dataset_name}/{method}/{run}/sample1', metadata)

    tables, metadata = remove_sdv_columns(tables, metadata)
    tables_synthetic, metadata = remove_sdv_columns(tables_synthetic, metadata, update_metadata=False)

    tables_test = load_tables(f'{PROJECT_PATH}/data/original/{dataset_name.split("_")[0]}/', metadata)
    tables_test, _ = remove_sdv_columns(tables_test, metadata, update_metadata=False)
    # split the test data
    min_date = datetime.strptime('2014-10-01', '%Y-%m-%d')
    max_date = datetime.strptime('2014-11-01', '%Y-%m-%d')
    tables_test['historical'] = tables_test['historical'][(tables_test['historical']['Date'] >= min_date) & (tables_test['historical']['Date'] < max_date)]
    
    return tables, tables_synthetic, tables_test, metadata


def load_airbnb(method):
    dataset_name = 'airbnb-simplified_subsampled'
    run = "1"
    metadata = Metadata().load_from_json(f'{PROJECT_PATH}/data/original/{dataset_name}/metadata.json')

    tables = load_tables(f'{PROJECT_PATH}/data/original/{dataset_name}/', metadata)
    tables_synthetic = load_tables(f'{PROJECT_PATH}/data/synthetic/{dataset_name}/{method}/{run}/sample1', metadata)

    tables, metadata = remove_sdv_columns(tables, metadata)
    for table in tables:
        if 'index' in tables_synthetic[table].columns:
            tables_synthetic[table].drop(columns=['index'], inplace=True)
    tables_synthetic, metadata = remove_sdv_columns(tables_synthetic, metadata, update_metadata=False)

    tables_test = load_tables(f'{PROJECT_PATH}/data/original/{dataset_name.split("_")[0]}/', metadata)
    tables_test, _ = remove_sdv_columns(tables_test, metadata, update_metadata=False)

    # select users with at most 50 sessions
    sessions_count = tables_test['sessions'].reset_index().groupby('user_id').index.count()
    eligable_users = sessions_count[sessions_count <= 50].index 
    no_sessions = tables_test['users'][~tables_test['users']['id'].isin(tables_test['sessions']['user_id'])]['id']
    eligable_users = eligable_users.union(no_sessions)
    # select users that are not in the synthetic data
    eligable_users = eligable_users[~eligable_users.isin(tables['users']['id'])]

    selected_users = np.random.choice(eligable_users, 2000, replace=False)
    tables_test['users'] = tables_test['users'][tables_test['users']['id'].isin(selected_users)]
    tables_test['sessions'] = tables_test['sessions'][tables_test['sessions']['user_id'].isin(selected_users)]

    return tables, tables_synthetic, tables_test, metadata


def load_walmart(method):
    dataset_name = 'walmart_subsampled_12'
    run = "1"
    metadata = Metadata().load_from_json(f'{PROJECT_PATH}/data/original/{dataset_name}/metadata.json')

    tables = load_tables(f'{PROJECT_PATH}/data/original/{dataset_name}/', metadata)
    tables_synthetic = load_tables(f'{PROJECT_PATH}/data/synthetic/{dataset_name}/{method}/{run}/sample1', metadata)

    tables, metadata = remove_sdv_columns(tables, metadata)
    tables_synthetic, metadata = remove_sdv_columns(tables_synthetic, metadata, update_metadata=False)

    # split the test data
    tables_test = load_tables(f'{PROJECT_PATH}/data/original/{dataset_name.split("_")[0]}/', metadata)
    tables_test, _ = remove_sdv_columns(tables_test, metadata, update_metadata=False)

    min_date = datetime.strptime('2012-01-01', '%Y-%m-%d')
    max_date = datetime.strptime('2012-02-01', '%Y-%m-%d')
    tables_test['depts'] = tables_test['depts'][(tables_test['depts']['Date'] >= min_date) & (tables_test['depts']['Date'] < max_date)]
    tables_test['features'] = tables_test['features'][(tables_test['features']['Date'] >= min_date) & (tables_test['features']['Date'] < max_date)]

    return tables, tables_synthetic, tables_test, metadata

## DATA PREPARATION
def process_rossmann(tables, metadata):
    df = tables['historical'].merge(tables['store'], on='Store')
    numerical_columns = [] 
    for table in metadata.get_tables():
        table_metadata = metadata.get_table_meta(table)
        for column, column_info in table_metadata['columns'].items():
            if column_info['sdtype'] == 'numerical':
                numerical_columns.append(column)
            elif column_info['sdtype'] == 'id':
                if column in df.columns and column != 'Store':
                    df.drop(columns=[column], inplace=True)

    # drop the StateHoliday column as it is constant causing problems with standardization
    # drop the DayOfWeek as we will be aggregating the data by month
    df.drop(columns=['StateHoliday', 'DayOfWeek'], inplace=True)

    # set the categories for categorical variables (in case methods do not generate certain values)
    df['Open'] = pd.Categorical(df['Open'], categories=['0', '1'])
    df['Promo'] = pd.Categorical(df['Promo'], categories=['0', '1'])
    df['SchoolHoliday'] = pd.Categorical(df['SchoolHoliday'], categories=['0', '1'])
    df['StoreType'] = pd.Categorical(df['StoreType'], categories=['a', 'b', 'c', 'd'])
    df['Assortment'] = pd.Categorical(df['Assortment'], categories=['a', 'b', 'c'])
    df['PromoInterval'] = pd.Categorical(df['PromoInterval'].astype('object').fillna('missing'), categories=['missing', 'Jan,Apr,Jul,Oct', 'Mar,Jun,Sept,Dec', 'Feb,May,Aug,Nov'])
    # one-hot encode the categorical columns
    df = pd.get_dummies(df, columns=['Open', 'Promo', 'SchoolHoliday', 'StoreType', 'Assortment', 'PromoInterval'], drop_first=True)
    # aggregate the data by store and month
    df['Month'] = df['Date'].dt.month
    df['Year'] = df['Date'].dt.year
    df = df.groupby(['Store', 'Month', 'Year']).mean().reset_index(drop=True)
    # drop the dates due to subsampling
    df.drop(columns=['Date'], inplace=True)

    df[numerical_columns] = df[numerical_columns].fillna(0)
    y = df.pop('Customers')

    # remove missing / infinite y values
    mask = y.isna() | y.isin([np.inf, -np.inf])
    X = df[~mask]
    y = y[~mask]
    return X, y


def process_airbnb(tables, metadata, categories):
    df = tables['users'].copy()
    df.drop(columns=['date_first_booking'], inplace=True)  

    for table in metadata.get_tables():
        table_metadata = metadata.get_table_meta(table)
        for column, column_info in table_metadata['columns'].items():
            if column_info['sdtype'] == 'categorical':
                if column in df.columns:
                    df[column] = pd.Categorical(df[column], categories=categories[column])
    
    # impute missing values 
    df[['age']] = df[['age']].fillna(0)
    
    # aggregate the sessions data by user_id to obtain average session duration
    sessions_data = tables['sessions'][['secs_elapsed', 'user_id']].groupby('user_id').mean()
    # add the sessions count
    sessions_count = tables['sessions']['user_id'].value_counts().rename('sessions_count')
    # join the sessions data and count of sessions
    sessions_data = sessions_data.join(sessions_count, on='user_id')

    # merge the sessions data with the users data and fill missing values with 0
    df = df.merge(sessions_data, left_on='id', right_index=True, how='left')
    df[sessions_data.columns] = df[sessions_data.columns].fillna(0)
    df.drop(columns=['id'], inplace=True)
    
    y = df.pop('country_destination')
    X = df.copy()
    
    # convert y to binary variable determining if the user booked a trip or not
    y = y != 'NDF'
        
    return X, y


def process_walmart(tables, metadata):
    df = tables['depts'].merge(tables['stores'], on='Store').merge(tables['features'], on=['Store', 'Date'], suffixes=('', '_y'))
    df.drop(df.filter(regex='_y$').columns, axis=1, inplace=True)
    df.drop(columns=['Dept'], inplace=True)
    

    # one-hot encode the categorical columns
    df['Type'] = pd.Categorical(df['Type'], categories=['A', 'B', 'C'])
    df = pd.get_dummies(df, columns=['Type'])
    # obtain average daily sales across all departments 
    df = df.groupby(['Store', 'Date']).mean().reset_index(drop=True)
    
    y = df.pop('Weekly_Sales')

    # remove missing / infinite y values
    mask = y.isna() | y.isin([np.inf, -np.inf])
    X = df[~mask]
    y = y[~mask]
    return X, y


## UTILITY
def load_dataset(dataset_name, method):
    if dataset_name == 'rossmann':
        tables, tables_synthetic, tables_test, metadata = load_rossmann(method)
        feature_engineering_function = process_rossmann
        target=('historical', 'Customers', 'regression')
    elif dataset_name == 'airbnb':
        tables, tables_synthetic, tables_test, metadata = load_airbnb(method)
        # Store the categories from the real data to avoid mistakes in case methods do not generate certain values
        categories = {}
        for _, table in tables_test.items():
            for column in table.columns:
                if table[column].dtype.name == 'category':
                    categories[column] = table[column].cat.categories.tolist()
        feature_engineering_function = partial(process_airbnb, categories=categories)
        # feature_engineering_function = None
        target=('users', 'country_destination', 'classification')
    elif dataset_name == 'walmart':
        tables, tables_synthetic, tables_test, metadata = load_walmart(method)
        feature_engineering_function = process_walmart
        target=('depts', 'Weekly_Sales', 'regression')
    else:
        raise ValueError(f"Dataset {dataset_name} not supported.")
    
    return tables, tables_synthetic, tables_test, metadata, feature_engineering_function, target


methods = [
    'COPY',
    'SDV', 
    'RCTGAN',
    'REALTABFORMER',
    'MOSTLYAI',
    'GRETEL_ACTGAN',
    'GRETEL_LSTM'
]
    
datasets = [
    'rossmann',
    'airbnb',
    'walmart',
]

classifiers = {
    'regression': {
        'xgboost': xgb.XGBRegressor,
        'linear': LinearRegression,
        'random_forest': RandomForestRegressor,
        'decision_tree': DecisionTreeRegressor,
        'knn': KNeighborsRegressor,
        'svr': SVR,
        'mlp': MLPRegressor,
    },
    'classification': {
        'xgboost': xgb.XGBClassifier,
        'linear': LogisticRegression,
        'random_forest': RandomForestClassifier,
        'decision_tree': DecisionTreeClassifier,
        'knn': KNeighborsClassifier,
        'svc': SVC,
        'gaussian_nb': GaussianNB,
        'mlp': MLPClassifier,
    }
}

cls_args = {
    'xgboost': {'random_state': None, 'seed': None},
    'linear': {},
    'random_forest': {'random_state': None},
    'decision_tree': {'random_state': None},
    'knn': {},
    'svr': {},
    'svc': {'random_state': None, 'probability': True},
    'gaussian_nb': {},
    'mlp': {'random_state': None},
}


feature_selection_models = {
    'regression': ['xgboost'],
    'classification': ['xgboost']
}

if __name__ == '__main__':
    from dotenv import load_dotenv

    load_dotenv()

    PROJECT_PATH = os.getenv("PROJECT_PATH")
    RESULTS_PATH = os.getenv("RESULTS_PATH")
    
    args = argparse.ArgumentParser()
    args.add_argument("--dataset-name", type=str, default="rossmann", choices=datasets, help="Dataset name to run the experiment on.")
    args.add_argument("--methods", '-m', default=None, help="List of methods to evaluate.", action='append')
    args.add_argument("--seed", type=int, default=0, help="Seed for reproducibility.")
    args.add_argument("--bootstrap_repetitions", type=int, default=100, help="Number of bootstrap samples.")
    args = args.parse_args()

    dataset_name = args.dataset_name
    if args.methods is None:
        methods_to_run = methods
    else:
        methods_to_run = args.methods
    seed = args.seed
    m = args.bootstrap_repetitions

    results = {}
    results[dataset_name] = {}
    for method in methods_to_run:
        print(f"Method: {method}, Dataset: {dataset_name}")
        results[dataset_name][method] = {}
        tables, tables_synthetic, tables_test, metadata, feature_engineering_function, target = load_dataset(dataset_name, method)
        task = target[2]
        for classifier, classifier_cls in classifiers[task].items():
            classifier_args_ = cls_args[classifier]
            if 'random_state' in classifier_args_:
                classifier_args_['random_state'] = seed
            if 'seed' in classifier_args_:
                classifier_args_['seed'] = seed

            ml_metric = MachineLearningEfficacyMetric(target=target, feature_engineering_function=feature_engineering_function,
                                                    classifier_cls=classifier_cls, random_state=seed, classifier_args=classifier_args_)
            

            result = ml_metric.run(tables, tables_synthetic, metadata, tables_test, m=m, feature_importance=classifier in feature_selection_models[task])
            print(f"Classifier: {classifier} real_score: {result['real_score'] :.3f} +- {result['real_score_se']:.3f}, synthetic_score: {result['synthetic_score']:.3f} +- {result['synthetic_score_se']:.3f}")
            importances_real = result.pop('importance_real', None)
            importances_syn = result.pop('importance_synthetic', None)
            results[dataset_name][method][classifier] = result
            if classifier in feature_selection_models[task]:
                feature_importances_spearman = []
                feature_importances_tau = []
                feature_importances_weighted = []
                feature_names = np.array(result['feature_names'])

                true_feature_importance_real = np.array(result['true_feature_importance_real'])
                true_feature_importance_synthetic = np.array(result['true_feature_importance_synthetic'])
                real_rank = np.argsort(true_feature_importance_real)
                synthetic_rank = np.argsort(true_feature_importance_synthetic)
                true_features_spearman_rank = spearmanr(feature_names[real_rank], feature_names[synthetic_rank]).statistic
                real_feature_order = feature_names[real_rank]
                synthetic_feature_order = feature_names[synthetic_rank]
                true_features_tau_rank = kendalltau(feature_names[real_rank], feature_names[synthetic_rank]).statistic
                indexed_real_rank = np.arange(len(real_rank))
                indexed_synthetic_rank = np.array([real_rank.tolist().index(feature) for feature in synthetic_rank])
                true_features_weighted_rank = weightedtau(indexed_real_rank, indexed_synthetic_rank, rank=None).statistic
                
                for i in range(m):
                    # importance_real = importances_real[i]
                    importance_syn = importances_syn[i]
                    # real_rank = np.argsort(importance_real)
                    synthetic_rank = np.argsort(importance_syn)
                    features_spearman_rank = spearmanr(feature_names[real_rank], feature_names[synthetic_rank]).statistic
                    features_tau_rank = kendalltau(feature_names[real_rank], feature_names[synthetic_rank]).statistic
                    
                    indexed_real_rank = np.arange(len(real_rank))
                    indexed_synthetic_rank_boot = np.array([real_rank.tolist().index(feature) for feature in synthetic_rank])
                    
                    features_weighted_rank = weightedtau(indexed_real_rank, indexed_synthetic_rank_boot, rank=None).statistic
                    feature_importances_spearman.append(features_spearman_rank)
                    feature_importances_tau.append(features_tau_rank)   
                    feature_importances_weighted.append(features_weighted_rank)

        
        # rank the classifiers
        real_classifier_rank = list(dict(sorted(results[dataset_name][method].items(), key=lambda x: x[1]['real_score'], reverse=True)).keys())
        synthetic_classifier_rank = list(dict(sorted(results[dataset_name][method].items(), key=lambda x: x[1]['synthetic_score'], reverse=True)).keys())
        print(f"Real data classifier rank: {real_classifier_rank}")
        print(f"Synthetic data classifier rank: {synthetic_classifier_rank}")

        true_classifier_rank_spearman = spearmanr(real_classifier_rank, synthetic_classifier_rank).statistic
        true_classifier_rank_kendall = kendalltau(real_classifier_rank, synthetic_classifier_rank).statistic

        indexed_real_rank = np.array([len(real_classifier_rank) - real_classifier_rank.index(classifier) for classifier in real_classifier_rank])
        indexed_synthetic_rank = np.array([len(real_classifier_rank) - real_classifier_rank.index(classifier) for classifier in synthetic_classifier_rank])
        true_classifier_rank_weighted = weightedtau(indexed_real_rank, indexed_synthetic_rank, rank=None).statistic


        classifier_rank_array_spearman = []
        classifier_rank_array_kendall = []
        classifier_rank_array_weighted = []
            
        for bootstrap_index in range(m):
            synthetic_classifier_rank_boot = list(dict(sorted(results[dataset_name][method].items(), key=lambda x: x[1]['synthetic_score_array'][bootstrap_index], reverse = True)).keys())

            classifier_rank_array_spearman.append(spearmanr(real_classifier_rank, synthetic_classifier_rank_boot).statistic)
            classifier_rank_array_kendall.append(kendalltau(real_classifier_rank, synthetic_classifier_rank_boot).statistic)

            indexed_real_rank_boot = np.array([len(real_classifier_rank) - real_classifier_rank.index(classifier) for classifier in real_classifier_rank])
            indexed_synthetic_rank_boot = np.array([len(real_classifier_rank) - real_classifier_rank.index(classifier) for classifier in synthetic_classifier_rank_boot])
            classifier_rank_array_weighted.append(weightedtau(indexed_real_rank_boot, indexed_synthetic_rank_boot, rank=None).statistic)

        spearman_rank = spearmanr(list(real_classifier_rank), list(synthetic_classifier_rank))
        results[dataset_name][method]['classifier_rank'] = spearman_rank.statistic
        results[dataset_name][method]['feature_importance_spearman'] = true_features_spearman_rank
        results[dataset_name][method]['feature_importance_spearman_mean'] = np.mean(feature_importances_spearman)
        results[dataset_name][method]['feature_importance_spearman_se'] = np.std(feature_importances_spearman) / np.sqrt(m)
        results[dataset_name][method]['feature_importance_tau'] = true_features_tau_rank
        results[dataset_name][method]['feature_importance_tau_mean'] = np.mean(feature_importances_tau)
        results[dataset_name][method]['feature_importance_tau_se'] = np.std(feature_importances_tau) / np.sqrt(m)
        results[dataset_name][method]['feature_importance_weighted'] = true_features_weighted_rank
        results[dataset_name][method]['feature_importance_weighted_mean'] = np.mean(feature_importances_weighted)
        results[dataset_name][method]['feature_importance_weighted_se'] = np.std(feature_importances_weighted) / np.sqrt(m)
        results[dataset_name][method]['spearman'] = true_classifier_rank_spearman
        results[dataset_name][method]['spearman_mean'] = np.mean(classifier_rank_array_spearman) 
        results[dataset_name][method]['spearman_se'] = np.std(classifier_rank_array_spearman) / np.sqrt(m)
        results[dataset_name][method]['kendall'] = true_classifier_rank_kendall
        results[dataset_name][method]['kendall_mean'] = np.mean(classifier_rank_array_kendall)
        results[dataset_name][method]['kendall_se'] = np.std(classifier_rank_array_kendall) / np.sqrt(m)
        results[dataset_name][method]['weighted'] = true_classifier_rank_weighted
        results[dataset_name][method]['weighted_mean'] = np.mean(classifier_rank_array_weighted)
        results[dataset_name][method]['weighted_se'] = np.std(classifier_rank_array_weighted) / np.sqrt(m)
        results[dataset_name][method]['real_feature_order'] = real_feature_order
        results[dataset_name][method]['synthetic_feature_order'] = synthetic_feature_order


        print()
        print(f"Boot spearman: {np.mean(classifier_rank_array_spearman):.3f}+-{np.std(classifier_rank_array_spearman) / np.sqrt(m):.4f}")
        print(f"Boot kendall: {np.mean(classifier_rank_array_kendall) :.3f}+-{np.std(classifier_rank_array_kendall) / np.sqrt(m):.4f}")
        print(f"Boot weighted: {np.mean(classifier_rank_array_weighted) :.3f}+-{np.std(classifier_rank_array_weighted) / np.sqrt(m):.4f}")
        
        print(f"Feature importance spearman: {np.mean(feature_importances_spearman) :.3f}+-{np.std(feature_importances_spearman) / np.sqrt(m):.4f}")
        print(f"Feature importance tau: {np.mean(feature_importances_tau) :.3f}+-{np.std(feature_importances_tau) / np.sqrt(m):.4f}")
        print(f"Feature importance weighted: {np.mean(feature_importances_weighted) :.3f}+-{np.std(feature_importances_weighted) / np.sqrt(m):.4f}")
        print()

        if len(methods_to_run) < len(methods):
            with open(f'{RESULTS_PATH}/mle_{dataset_name}_{seed}_{method}.json', 'w') as f:
                json.dump(results, f, indent=4, cls=NpEncoder)
            
    with open(f'{RESULTS_PATH}/mle_{dataset_name}_{seed}.json', 'w') as f:
        json.dump(results, f, indent=4, cls=NpEncoder)



import os
import json
import argparse
from pathlib import Path

from syntherela.metadata import Metadata
from syntherela.data import load_tables, remove_sdv_columns

from ClavaDDPM.preprocess_utils import (
    encode_and_save,
    calculate_days_since_earliest_date,
)


def remap_ids(metadata, tables):
    # Transform the ids to 0, 1, 2, ...
    id_map = {}

    for parent_table_name in metadata.get_tables():
        primary_key = f"{parent_table_name}_id"

        if parent_table_name not in id_map:
            id_map[parent_table_name] = {}

        if primary_key not in id_map[parent_table_name]:
            id_map[parent_table_name][primary_key] = {}
            idx = 0
            for primary_key_val in tables[parent_table_name][primary_key].unique():
                id_map[parent_table_name][primary_key][primary_key_val] = idx
                idx += 1

        for relationship in metadata.relationships:
            if relationship["parent_table_name"] != parent_table_name:
                continue
            if relationship["child_table_name"] not in id_map:
                id_map[relationship["child_table_name"]] = {}
            fk = f"{parent_table_name}_id"
            id_map[relationship["child_table_name"]][fk] = id_map[parent_table_name][fk]

    # remap the ids
    for table_name in id_map.keys():
        for column_name in id_map[table_name].keys():
            if column_name not in tables[table_name].columns:
                raise ValueError(
                    f"Column {column_name} not found in table {table_name}"
                )
            tables[table_name][column_name] = tables[table_name][column_name].map(
                id_map[table_name][column_name]
            )

    return tables


def rename_ids(metadata, tables):
    non_integer_ids = False
    ids = dict()
    for table_name in metadata.get_tables():
        ids[table_name] = []
        primary_key = metadata.get_primary_key(table_name)
        if primary_key is not None:
            tables[table_name][f"{table_name}_id"] = tables[table_name][primary_key]
            pk = tables[table_name].pop(primary_key)
            ids[table_name].append(f"{table_name}_id")
            if pk.dtype != "int64":
                non_integer_ids = True
        else:
            tables[table_name][f"{table_name}_id"] = range(len(tables[table_name]))
            ids[table_name].append(f"{table_name}_id")
        for parent in metadata.get_parents(table_name):
            foreign_keys = metadata.get_foreign_keys(parent, table_name)
            assert (
                len(foreign_keys) == 1
            ), "CLAVADDPM only supports one foreign key per table pair."
            foreign_key = foreign_keys[0]
            tables[table_name][f"{parent}_id"] = tables[table_name][foreign_key]
            fk = tables[table_name].pop(foreign_key)
            ids[table_name].append(f"{parent}_id")
            if fk.dtype != "int64":
                non_integer_ids = True

    if non_integer_ids:
        tables = remap_ids(metadata, tables)
    return tables, ids


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("dataset_name", type=str, help="Name of the dataset")

    parser.add_argument(
        "--num_clusters",
        type=int,
        default=50,
        help="Number of clusters to use for clustering.",
    )
    parser.add_argument(
        "--real-data-path",
        type=str,
        default="data/original/",
        help="Path to the original data.",
    )
    parser.add_argument("--run-id", type=str, default="1")

    return parser.parse_args()


def main(args):
    num_clusters = args.num_clusters
    dataset_name = args.dataset_name
    real_data_path = args.real_data_path
    run_id = 1

    metadata = Metadata().load_from_json(
        Path(real_data_path) / f"{dataset_name}/metadata.json"
    )
    real_data = load_tables(Path(real_data_path) / f"{dataset_name}", metadata)
    real_data, metadata = remove_sdv_columns(real_data, metadata)

    processed_data_path = os.path.join("complex_data", dataset_name)
    os.makedirs(processed_data_path, exist_ok=True)

    tables = dict()
    relation_order = []
    for table in metadata.get_tables():
        parents = metadata.get_parents(table)
        children = metadata.get_children(table)
        tables[table] = {
            "children": list(children),
            "parents": list(parents),
        }
        if not len(parents):
            relation_order.append([None, table])

    for relation in metadata.relationships:
        parent = relation["parent_table_name"]
        child = relation["child_table_name"]
        relation_order.append([parent, child])

    dataset_meta = {
        "relation_order": relation_order,
        "tables": tables,
    }

    # Create a dataset_meta.json, in which tables should be manually created to
    #    specify all foreign key relationships in a multi-table dataset.
    with open(os.path.join(processed_data_path, "dataset_meta.json"), "w") as f:
        json.dump(dataset_meta, f, indent=4)

    first_dates = dict()

    # Save all tables as .csv files. All id columns should be named <column_name>_id.
    real_data, ids = rename_ids(metadata, real_data)
    for table_name, df in real_data.items():
        id_columns = ids[table_name]
        categorical_columns = metadata.get_column_names(
            table_name, sdtype="categorical"
        )
        datetime_columns = metadata.get_column_names(table_name, sdtype="datetime")
        numerical_columns = metadata.get_column_names(table_name, sdtype="numerical")

        for col in numerical_columns:
            df[col] = df[col].fillna(df[col].mean())

        # Calculate days since the earliest date for each datetime column
        if len(datetime_columns) > 0:
            first_dates[table_name] = dict()
        for col in datetime_columns:
            # TODO: handle missing values in datetime columns
            if df[col].isnull().sum() > 0:
                df[col] = df[col].fillna(df[col].mode().iloc[0])
            df[col], first_dates[table_name][col] = calculate_days_since_earliest_date(
                df[col].dt.strftime("%y%m%d")
            )

        # Create a domain file for each table, id columns excluded.
        encode_and_save(
            df,
            categorical_columns,
            id_columns,
            processed_data_path,
            table_name,
        )

    if len(first_dates) > 0:
        with open(os.path.join(processed_data_path, "first_dates.json"), "w") as f:
            json.dump(first_dates, f, indent=4)

    # create config.json
    config = {
        "general": {
            "data_dir": processed_data_path.replace("\\", "/"),
            "exp_name": f"{dataset_name}_train",
            "workspace_dir": f"clavaDDPM_workspace_run{run_id}/{dataset_name}",
            "sample_prefix": "",
            "test_data_dir": None,
        },
        "clustering": {
            "parent_scale": 1.0,
            "num_clusters": num_clusters,
            "clustering_method": "both",
        },
        "diffusion": {
            "d_layers": [512, 1024, 1024, 1024, 1024, 512],
            "dropout": 0.0,
            "num_timesteps": 2000,
            "model_type": "mlp",
            "iterations": 200000,
            "batch_size": 4096,
            "lr": 6e-4,
            "gaussian_loss_type": "mse",
            "weight_decay": 1e-5,
            "scheduler": "cosine",
        },
        "classifier": {
            "d_layers": [128, 256, 512, 1024, 512, 256, 128],
            "lr": 0.0001,
            "dim_t": 128,
            "batch_size": 4096,
            "iterations": 20000,
        },
        "sampling": {"batch_size": 20000, "classifier_scale": 1.0},
        "matching": {
            "num_matching_clusters": 1,
            "matching_batch_size": 1000,
            "unique_matching": True,
            "no_matching": False,
        },
    }

    with open(
        os.path.join(
            "ClaVaDDPM",
            "configs",
            f"{dataset_name}_run{run_id}.json",
        ),
        "w",
    ) as f:
        json.dump(config, f, indent=4)


if __name__ == "__main__":
    args = parse_arguments()
    main(args)



# -*- coding: utf-8 -*-

"""Top-level package for rdt2."""


__author__ = """MIT Data To AI Lab"""
__email__ = 'dailabmit@gmail.com'
__version__ = '1.0.0.dev1'

import numpy as np
import pandas as pd

from rctgan.rdt2 import transformers
from rctgan.rdt2.hyper_transformer import HyperTransformer

__all__ = [
    'HyperTransformer',
    'transformers'
]

RANDOM_SEED = 42


def get_demo(num_rows=5):
    """Generate demo data with multiple sdtypes.

    The first five rows are hard coded. The rest are randomly generated
    using ``np.random.seed(42)``.

    Args:
        num_rows (int):
            Number of data rows to generate. Defaults to 5.

    Returns:
        pd.DataFrame
    """
    # Hard code first five rows
    login_dates = ['2021-06-26', '2021-02-10', 'NAT', '2020-09-26', '2020-12-22']
    last_login = [np.datetime64(i) for i in login_dates]
    email_optin = pd.Series([False, False, False, True, np.nan], dtype='object')
    credit_card = ['VISA', 'VISA', 'AMEX', np.nan, 'DISCOVER']
    age = [29, 18, 21, 45, 32]
    dollars_spent = [99.99, np.nan, 2.50, 25.00, 19.99]

    data = pd.DataFrame({
        'last_login': last_login,
        'email_optin': email_optin,
        'credit_card': credit_card,
        'age': age,
        'dollars_spent': dollars_spent
    })

    if num_rows <= 5:
        return data.iloc[:num_rows]

    # Randomly generate the remaining rows
    random_state = np.random.get_state()
    np.random.set_state(np.random.RandomState(RANDOM_SEED).get_state())
    try:
        num_rows -= 5

        login_dates = np.array([
            np.datetime64('2000-01-01') + np.timedelta64(np.random.randint(0, 10000), 'D')
            for _ in range(num_rows)
        ])
        login_dates[np.random.random(size=num_rows) > 0.8] = np.datetime64('NaT')

        email_optin = pd.Series([True, False, np.nan], dtype='object').sample(
            num_rows, replace=True)
        credit_card = np.random.choice(['VISA', 'AMEX', np.nan, 'DISCOVER'], size=num_rows)
        age = np.random.randint(18, 100, size=num_rows)

        dollars_spent = np.around(np.random.uniform(0, 100, size=num_rows), decimals=2)
        dollars_spent[np.random.random(size=num_rows) > 0.8] = np.nan

    finally:
        np.random.set_state(random_state)

    return data.append(pd.DataFrame({
        'last_login': login_dates,
        'email_optin': email_optin,
        'credit_card': credit_card,
        'age': age,
        'dollars_spent': dollars_spent
    }), ignore_index=True)



import xgboost as xgb
import seaborn as sns
from matplotlib import rc
import matplotlib.pyplot as plt

from syntherela.metadata import Metadata
from syntherela.metrics.multi_table.detection import AggregationDetection
from syntherela.data import load_tables, remove_sdv_columns

sns.set_theme()
rc("font", **{"family": "serif", "serif": ["Times"], "size": 30})
rc("text", usetex=True)


def reproduce_figure(tables, tables_synthetic, metadata, dataset_name, figure_name):
    # Compute the metric
    xgb_cls = xgb.XGBClassifier
    xgb_args = {
        "seed": 0,
        "importance_type": "gain",
    }

    metric = AggregationDetection(
        classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=42
    )

    for table in tables.keys():
        tables_synthetic[table] = tables_synthetic[table][tables[table].columns]

    if dataset_name == "imdb_MovieLens_v1":
        target_table = "movies"
    elif dataset_name == "Biodegradability_v1":
        target_table = "molecule"

    metric.run(
        tables,
        tables_synthetic,
        metadata=metadata,
        target_table=target_table,
    )

    # Plot the feature importance

    fig, ax = plt.subplots(figsize=(7, 7))
    metric.plot_feature_importance(metadata, ax=ax, combine_categorical=True)
    plt.savefig(
        f"results/figures/figure4{figure_name}.png", bbox_inches="tight", dpi=600
    )


## FIGURE 4 (a)
dataset_name = "Biodegradability_v1"
method = "GRETEL_LSTM"

metadata = Metadata().load_from_json(f"data/original/{dataset_name}/metadata.json")

tables = load_tables(f"data/original/{dataset_name}/", metadata)
tables_synthetic = load_tables(
    f"data/synthetic/{dataset_name}/{method}/1/sample1", metadata
)

tables, metadata = remove_sdv_columns(tables, metadata)
tables_synthetic, metadata = remove_sdv_columns(
    tables_synthetic, metadata, update_metadata=False
)

reproduce_figure(tables, tables_synthetic, metadata, dataset_name, "a")

## FIGURE 4 (a)
dataset_name = "imdb_MovieLens_v1"
method = "CLAVADDPM"

metadata = Metadata().load_from_json(f"data/original/{dataset_name}/metadata.json")

tables = load_tables(f"data/original/{dataset_name}/", metadata)
tables_synthetic = load_tables(
    f"data/synthetic/{dataset_name}/{method}/1/sample1", metadata
)

tables, metadata = remove_sdv_columns(tables, metadata)
tables_synthetic, metadata = remove_sdv_columns(
    tables_synthetic, metadata, update_metadata=False
)

reproduce_figure(tables, tables_synthetic, metadata, dataset_name, "b")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as readme_file:
    readme = readme_file.read()


install_requires = [
    'Faker>=3.0.0,<10',
    'graphviz>=0.13.2,<1',
    "numpy>=1.18.0,<1.20.0;python_version<'3.7'",
    "numpy>=1.20.0,<2;python_version>='3.7'",
    'pandas>=1.1.3,<2',
    'tqdm>=4.15,<5',
    'rdt>=0.6.1,<0.7',
    'sdmetrics>=0.4.1,<0.5',
    'tqdm>=4.63.0,<4.64.1'
]

pomegranate_requires = [
    "pomegranate>=0.13.4,<0.14.2;python_version<'3.7'",
    "pomegranate>=0.14.1,<0.15;python_version>='3.7'",
]

setup_requires = [
    'pytest-runner>=2.11.1',
]

tests_require = [
    'pytest>=3.4.2',
    'pytest-cov>=2.6.0',
    'pytest-rerunfailures>10',
    'jupyter>=1.0.0,<2',
    'rundoc>=0.4.3,<0.5',
]

development_requires = [
    # general
    'bumpversion>=0.5.3,<0.6',
    'pip>=9.0.1',
    'watchdog>=0.8.3,<0.11',

    # docs
    'docutils>=0.12,<0.18',
    'm2r2>=0.2.5,<0.3',
    'nbsphinx>=0.5.0,<0.7',
    'Sphinx>=3,<3.3',
    'pydata-sphinx-theme<0.5',

    # Jinja2>=3 makes the sphinx theme fail
    'Jinja2>=2,<3',

    # style check
    'flake8>=3.7.7,<4',
    'flake8-absolute-import>=1.0,<2',
    'flake8-docstrings>=1.5.0,<2',
    'flake8-sfs>=0.0.3,<0.1',
    'isort>=4.3.4,<5',

    # fix style issues
    'autoflake>=1.1,<2',
    'autopep8>=1.4.3,<1.6',

    # distribute on PyPI
    'twine>=1.10.0,<4',
    'wheel>=0.30.0',

    # Advanced testing
    'coverage>=4.5.1,<6',
    'tox>=2.9.1,<4',
    'invoke'
]

setup(
    author='Croesus Lab',
    author_email='mohamed.gueye@croesus.com',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description='Synthetic Data Generation for relational data.',
    extras_require={
        'test': tests_require,
        'dev': development_requires + tests_require,
        'pomegranate': pomegranate_requires,
    },
    include_package_data=True,
    install_requires=install_requires,
    keywords='rctgan synthetic-data synhtetic-data-generation multi-table',
    license='Apache license',
    long_description=readme + '\n\n',
    long_description_content_type='text/markdown',
    name='rctgan',
    packages=find_packages(include=['rctgan', 'rctgan.*']),
    python_requires='>=3.6,<3.10',
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://www.croesus.com/about-us/croesus-lab/',
    version='1.01',
    zip_safe=False,
)


"""Functions to load demo datasets."""

import io
import logging
import os
import urllib.request
from datetime import datetime, timedelta
from zipfile import ZipFile

import numpy as np
import pandas as pd
import scipy as sp
from faker import Faker

from rctgan.metadata import Metadata, Table

LOGGER = logging.getLogger(__name__)


DEMO_METADATA = {
    'tables': {
        'users': {
            'primary_key': 'user_id',
            'fields': {
                'user_id': {
                    'type': 'id',
                    'subtype': 'integer'
                },
                'country': {
                    'type': 'categorical'
                },
                'gender': {
                    'type': 'categorical'
                },
                'age': {
                    'type': 'numerical',
                    'subtype': 'integer'
                }
            }
        },
        'sessions': {
            'primary_key': 'session_id',
            'fields': {
                'session_id': {
                    'type': 'id',
                    'subtype': 'integer'
                },
                'user_id': {
                    'ref': {
                        'field': 'user_id',
                        'table': 'users'
                    },
                    'type': 'id',
                    'subtype': 'integer'
                },
                'device': {
                    'type': 'categorical'
                },
                'os': {
                    'type': 'categorical'
                },
                'minutes': {
                    'type': 'numerical',
                    'subtype': 'integer'
                }
            }
        },
        'transactions': {
            'primary_key': 'transaction_id',
            'fields': {
                'transaction_id': {
                    'type': 'id',
                    'subtype': 'integer'
                },
                'session_id': {
                    'ref': {
                        'field': 'session_id',
                        'table': 'sessions'
                    },
                    'type': 'id',
                    'subtype': 'integer'
                },
                'timestamp': {
                    'type': 'datetime',
                    'format': '%Y-%m-%dT%H:%M'
                },
                'amount': {
                    'type': 'numerical',
                    'subtype': 'float'
                },
                'cancelled': {
                    'type': 'boolean'
                }
            }
        }
    }
}


DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')
DATA_URL = 'https://sdv-datasets.s3.amazonaws.com/{}.zip'
DATASETS_URL = 'https://sdv-datasets.s3.amazonaws.com/datasets.csv'


def _dtypes64(table):
    for name, column in table.items():
        if column.dtype == np.int32:
            table[name] = column.astype('int64')
        elif column.dtype == np.float32:
            table[name] = column.astype('float64')

    return table


def _download(dataset_name, data_path):
    url = DATA_URL.format(dataset_name)

    LOGGER.info('Downloading dataset {} from {}'.format(dataset_name, url))
    response = urllib.request.urlopen(url)
    bytes_io = io.BytesIO(response.read())

    LOGGER.info('Extracting dataset into {}'.format(data_path))
    with ZipFile(bytes_io) as zf:
        zf.extractall(data_path)


def _get_dataset_path(dataset_name, data_path):
    if not os.path.exists(data_path):
        os.makedirs(data_path)

    if not os.path.exists(os.path.join(data_path, dataset_name)):
        _download(dataset_name, data_path)

    return os.path.join(data_path, dataset_name)


def _load_relational_dummy():
    users = pd.DataFrame({
        'user_id': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        'country': ['US', 'UK', 'ES', 'UK', 'US', 'DE', 'BG', 'ES', 'FR', 'UK'],
        'gender': ['M', 'F', None, 'M', 'F', 'M', 'F', None, 'F', None],
        'age': [34, 23, 44, 22, 54, 57, 45, 41, 23, 30]
    })
    sessions = pd.DataFrame({
        'session_id': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        'user_id': [0, 1, 1, 2, 4, 5, 6, 6, 6, 8],
        'device': ['mobile', 'tablet', 'tablet', 'mobile', 'mobile',
                   'mobile', 'mobile', 'tablet', 'mobile', 'tablet'],
        'os': ['android', 'ios', 'android', 'android', 'ios',
               'android', 'ios', 'ios', 'ios', 'ios'],
        'minutes': [23, 12, 8, 13, 9, 32, 7, 21, 29, 34],
    })
    transactions = pd.DataFrame({
        'transaction_id': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        'session_id': [0, 0, 1, 3, 5, 5, 7, 8, 9, 9],
        'timestamp': ['2019-01-01T12:34:32', '2019-01-01T12:42:21', '2019-01-07T17:23:11',
                      '2019-01-10T11:08:57', '2019-01-10T21:54:08', '2019-01-11T11:21:20',
                      '2019-01-22T14:44:10', '2019-01-23T10:14:09', '2019-01-27T16:09:17',
                      '2019-01-29T12:10:48'],
        'amount': [100.0, 55.3, 79.5, 112.1, 110.0, 76.3, 89.5, 132.1, 68.0, 99.9],
        'cancelled': [False, False, False, True, True, False, False, True, False, False],
    })
    transactions['timestamp'] = pd.to_datetime(transactions['timestamp'])

    tables = {
        'users': _dtypes64(users),
        'sessions': _dtypes64(sessions),
        'transactions': _dtypes64(transactions),
    }

    return Metadata(DEMO_METADATA), tables


def sample_relational_demo(size=30):
    """Sample demo data with the indicate number of rows in the parent table."""
    # Users
    faker = Faker()
    countries = [faker.country_code() for _ in range(5)]
    country = np.random.choice(countries, size=size)
    gender = np.random.choice(['F', 'M', None], p=[0.5, 0.4, 0.1], size=size)
    age = (
        sp.stats.truncnorm.rvs(-1.2, 1.5, loc=30, scale=10, size=size).astype(int)
        + 3 * (gender == 'M')
        + 3 * (country == countries[0]).astype(int)
    )
    num_sessions = (
        sp.stats.gamma.rvs(1, loc=0, scale=2, size=size)
        * (0.8 + 0.2 * (gender == 'F'))
    ).round().astype(int)

    users = pd.DataFrame({
        'country': country,
        'gender': gender,
        'age': age,
        'num_sessions': num_sessions
    })
    users.index.name = 'user_id'

    # Sessions
    sessions = pd.DataFrame()
    for user_id, user in users.iterrows():
        device_weights = [0.1, 0.4, 0.5] if user.gender == 'M' else [0.3, 0.4, 0.3]
        devices = np.random.choice(
            ['mobile', 'tablet', 'pc'],
            size=user.num_sessions,
            p=device_weights
        )
        os = []
        pc_weights = [0.6, 0.3, 0.1] if user.age > 30 else [0.2, 0.4, 0.4]
        pc_os = np.random.choice(['windows', 'macos', 'linux'], p=pc_weights)
        phone_weights = [0.7, 0.3] if user.age > 30 else [0.9, 0.1]
        phone_os = np.random.choice(['android', 'ios'], p=phone_weights)
        for device in devices:
            os.append(pc_os if device == 'pc' else phone_os)

        minutes = (
            sp.stats.truncnorm.rvs(-3, 3, loc=30, scale=10, size=user.num_sessions)
            * (1 + 0.1 * (user.gender == 'M'))
            * (1 + user.age / 100)
            * (1 + 0.1 * (devices == 'pc'))
        )
        num_transactions = (minutes / 10) * (0.5 + (user.gender == 'F'))

        sessions = sessions.append(pd.DataFrame({
            'user_id': np.full(user.num_sessions, int(user_id)),
            'device': devices,
            'os': os,
            'minutes': minutes.round().astype(int),
            'num_transactions': num_transactions.round().astype(int),
        }), ignore_index=True)

    sessions.index.name = 'session_id'
    del users['num_sessions']

    # Transactions
    transactions = pd.DataFrame()
    for session_id, session in sessions.iterrows():
        size = session.num_transactions
        if size:
            amount_base = sp.stats.truncnorm.rvs(-2, 4, loc=100, scale=50, size=size)
            is_apple = session['os'] in ('ios', 'macos')
            amount_modif = np.random.random(size) * 100 * is_apple
            amount = amount_base / np.random.randint(1, size + 1) + amount_modif

            seconds = np.random.randint(3600 * 24 * 365)
            start = datetime(2019, 1, 1) + timedelta(seconds=seconds)

            timestamp = sorted([
                start + timedelta(seconds=int(seconds))
                for seconds in np.random.randint(60 * session.minutes, size=size)
            ])
            cancelled = np.random.random(size=size) < (1 / (size * 2))
            transactions = transactions.append(pd.DataFrame({
                'session_id': np.full(session.num_transactions, int(session_id)),
                'timestamp': timestamp,
                'amount': amount.round(2),
                'cancelled': cancelled,
            }), ignore_index=True)

    transactions.index.name = 'transaction_id'
    del sessions['num_transactions']

    tables = {
        'users': _dtypes64(users.reset_index()),
        'sessions': _dtypes64(sessions.reset_index()),
        'transactions': _dtypes64(transactions.reset_index()),
    }
    return Metadata(DEMO_METADATA), tables


def _load_demo_dataset(dataset_name, data_path):
    dataset_path = _get_dataset_path(dataset_name, data_path)
    meta = Metadata(metadata=os.path.join(dataset_path, 'metadata.json'))
    tables = {
        name: _dtypes64(table)
        for name, table in meta.load_tables().items()
    }
    return meta, tables


def load_demo(dataset_name=None, data_path=DATA_PATH, metadata=False):
    """Load relational demo data.

    If a dataset name is given, it is downloaded from the sdv-datasets S3 bucket.
    Otherwise, a toy dataset with three simple tables is loaded:

        * users: user data including country, gender and age.
        * sessions: sessions data with a foreign key to user.
        * transactions: transactions data with a foreign key to sessions.

    If ``metadata`` is ``True``, the output will be a tuple with a ``Metadata``
    instance for the dataset and a ``tables`` dict that contains the tables loaded
    as ``pandas.DataFrames``.
    If ``metadata`` is ``False``, only the ``tables`` are returned.

    Args:
        dataset_name (str):
            Dataset name to be downloaded, if ``None`` use default demo data. Defaults to ``None``.
        data_path (str):
            Data path to save the dataset files, only used if dataset_name is provided.
            Defaults to ``DATA_PATH``.
        metadata (bool):
            If ``True`` return Metadata object. Defaults to ``False``.

    Returns:
        dict or tuple:
            If ``metadata`` is ``False`` return a ``dict`` with the tables data.
            If ``metadata`` is ``True`` return a ``tuple`` with Metadata and tables data.
    """
    if dataset_name:
        meta, tables = _load_demo_dataset(dataset_name, data_path)
    else:
        meta, tables = _load_relational_dummy()

    if metadata:
        return meta, tables

    return tables


def _load_tabular_dummy():
    """Load a dummy tabular demo dataframe."""
    age = np.random.randint(30, 50, 12)
    age_when_joined = age - np.random.randint(1, 10, 12)
    years_exp = np.random.randint(1, 6, 12)
    contractor = [0.0, 1.0, 0.0, 1.0, 0.0, 0.0] * 2

    is_contractor = np.array(contractor).astype(bool)
    salary = np.random.randint(60, 320, 12) * 500.
    bonus = np.random.randint(10, 50, 12) * 500.
    salary[is_contractor] = np.random.uniform(30000, 160000, 4).round(2)
    bonus[is_contractor] = np.random.uniform(5000, 25000, 4).round(2)

    return pd.DataFrame({
        'company': ['Pear', 'Pear', 'Glasses', 'Glasses', 'Cheerper', 'Cheerper'] * 2,
        'department': ['Sales', 'Design', 'AI', 'Search Engine', 'BigData', 'Support'] * 2,
        'employee_id': [1, 5, 1, 7, 6, 11, 28, 75, 33, 56, 42, 80],
        'age': age,
        'age_when_joined': age_when_joined,
        'years_in_the_company': age - age_when_joined,
        'salary': salary,
        'annual_bonus': bonus,
        'prior_years_experience': years_exp,
        'full_time': [1.0, 0.0, 1.0, 0.0, 0.0, 0.0] * 2,
        'part_time': [0.0, 0.0, 0.0, 0.0, 1.0, 1.0] * 2,
        'contractor': contractor
    })


def load_tabular_demo(dataset_name=None, table_name=None, data_path=DATA_PATH, metadata=False):
    """Load a tabular demo.

    If a dataset name is given, it is downloaded from the sdv-datasets S3 bucket.
    Otherwise, a toy dataset with a single table that contains data from a short fake
    collection of employees.

    If ``metadata`` is ``True``, the output will be a tuple with a ``Metadata``
    instance for the dataset and a ``pandas.DataFrame`` with the data from the table.
    If ``metadata`` is ``False``, only the ``pandas.DataFrame`` is returned.

    Args:
        dataset_name (str):
            Dataset name to be downloaded, if ``None`` use default demo data. Defaults to ``None``.
        table_name (str):
            If a table name is given, return this table from the indicated dataset.
            Otherwise, return the first one.
        data_path (str):
            Data path to save the dataset files, only used if dataset_name is provided.
            Defaults to ``DATA_PATH``.
        metadata (bool):
            If ``True`` also return a Table object. Defaults to ``False``.

    Returns:
        pandas.DataFrame or tuple:
            If ``metadata`` is ``False`` return a ``pandas.DataFrame`` with the tables data.
            If ``metadata`` is ``True`` return a ``tuple`` with a Table and the data.
    """
    if dataset_name:
        meta, tables = _load_demo_dataset(dataset_name, data_path)

        if table_name is None:
            table_name = meta.get_tables()[0]

        table = _dtypes64(tables[table_name])

        if metadata:
            return Table.from_dict(meta.get_table_meta(table_name)), table

        return table

    table = _dtypes64(_load_tabular_dummy())
    if metadata:
        table_meta = Table.from_dict({
            'fields': {
                'company': {'type': 'categorical'},
                'department': {'type': 'categorical'},
                'employee_id': {'type': 'numerical', 'subtype': 'integer'},
                'age': {'type': 'numerical', 'subtype': 'integer'},
                'age_when_joined': {'type': 'numerical', 'subtype': 'integer'},
                'years_in_the_company': {'type': 'numerical', 'subtype': 'integer'},
                'salary': {'type': 'numerical', 'subtype': 'float'},
                'annual_bonus': {'type': 'numerical', 'subtype': 'float'},
                'prior_years_experience': {'type': 'numerical', 'subtype': 'integer'}
            },
            'constraints': [
                {
                    'constraint': 'UniqueCombinations',
                    'columns': ['company', 'department'],
                },
                {
                    'constraint': 'GreaterThan',
                    'low': 'age_when_joined',
                    'high': 'age'
                },
                {
                    'constraint': 'GreaterThan',
                    'low': 30000,
                    'high': 'salary'
                },
                {
                    'constraint': 'Positive',
                    'high': 'prior_years_experience'
                },
                {
                    'constraint': 'Rounding',
                    'columns': 'salary',
                    'digits': 2
                }
            ],
            'model_kwargs': {}
        })
        return table_meta, table

    return table


def load_timeseries_demo(dataset_name=None, table_name=None, metadata=False):
    """Load a timeseries demo.

    If a dataset name is given, it is downloaded from the sdv-datasets S3 bucket.
    Otherwise, a the NASDAQ100_2019 dataset is loaded.

    If ``metadata`` is ``True``, the output will be a tuple with a ``Metadata``
    instance for the dataset and a ``pandas.DataFrame`` with the data from the table.
    If ``metadata`` is ``False``, only the ``pandas.DataFrame`` is returned.

    Args:
        dataset_name (str):
            Dataset name to be downloaded, if ``None`` use default dataset. Defaults to ``None``.
        table_name (str):
            If a table name is given, return this table from the indicated dataset.
            Otherwise, return the first one.
        data_path (str):
            Data path to save the dataset files, only used if dataset_name is provided.
            Defaults to ``DATA_PATH``.
        metadata (bool):
            If ``True`` also return a Table object. Defaults to ``False``.

    Returns:
        pandas.DataFrame or tuple:
            If ``metadata`` is ``False`` return a ``pandas.DataFrame`` with the tables data.
            If ``metadata`` is ``True`` return a ``tuple`` with a Table and the data.
    """
    dataset_name = dataset_name or 'nasdaq100_2019'
    return load_tabular_demo(dataset_name, table_name, data_path=DATA_PATH, metadata=metadata)


def get_available_demos():
    """Get available demos and information about them.

    Returns:
        pandas.DataFrame:
            Table with the available demos.
    """
    return pd.read_csv(DATASETS_URL)



import os
import sys
import logging
import argparse
from pathlib import Path


from sdv.multi_table import HMASynthesizer
from syntherela.metadata import Metadata
from syntherela.data import load_tables, save_tables, remove_sdv_columns

MODEL_NAME = "SDV"

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="airbnb-simplified_subsampled")
args.add_argument("--real-data-path", type=str, default="data/original")
args.add_argument("--synthetic-data-path", type=str, default="data/synthetic")
args.add_argument("--model-save-path", type=str, default="checkpoints")
args.add_argument("--run-id", type=str, default="1")
args = args.parse_args()

dataset_name = args.dataset_name
real_data_path = args.real_data_path
synthetic_data_path = args.synthetic_data_path
model_save_path = args.model_save_path
run_id = args.run_id

logger = logging.getLogger(f"{MODEL_NAME}_logger")

logger.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.info(f"START LOGGING DATASET {dataset_name} RUN {run_id}...")


logger.debug("Loading real data...")
metadata = Metadata().load_from_json(
    Path(real_data_path) / f"{dataset_name}/metadata.json"
)

if dataset_name == "Biodegradability_v1":
    metadata.update_column(
        "bond", "type", sdtype="numerical", computer_representation="Int64"
    )

real_data = load_tables(Path(real_data_path) / f"{dataset_name}", metadata)
real_data, metadata = remove_sdv_columns(real_data, metadata)
metadata.validate_data(real_data)
logger.debug("Real data loaded")


synthetic_data = {}

# GENERATE SYNTHETIC DATA ---------------------------------
model = HMASynthesizer(metadata=metadata)
model.fit(real_data)

# SAVE MODEL CHECKPOINT -----------------------------------
logger.debug("Saving model checkpoint...")
model_path = os.path.join(model_save_path, MODEL_NAME, dataset_name)
os.makedirs(model_path, exist_ok=True)
model_path = os.path.join(model_path, f"model_{run_id}.pkl")
model.save(model_path)

# SAMPLE AND SAVE DATA ------------------------------------
logger.info("Sampling and saving synthetic data...")
for i in range(1, 4):
    logger.debug(f"Sampling sample {i}")
    model.seed = i
    synthetic_data = model.sample()
    save_data_path = (
        Path(synthetic_data_path) / dataset_name / MODEL_NAME / run_id / f"sample{i}"
    )
    save_tables(synthetic_data, save_data_path)
    logger.debug(f"Done! Sample {i} saved!")


logger.info("COMPLETE GENERATION DONE.")



from scipy.stats import ks_2samp
from sdmetrics.utils import get_cardinality_distribution

from syntherela.metrics.base import StatisticalBaseMetric


class CardinalityShapeSimilarity(StatisticalBaseMetric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "CardinalityShapeSimilarity"

    def validate(self, real_data, synthetic_data):
        return sorted(real_data.keys()) == sorted(synthetic_data.keys())

    def run(self, real_data, synthetic_data, metadata, **kwargs):
        self.validate(real_data, synthetic_data)
        return self.compute(real_data, synthetic_data, metadata)

    @staticmethod
    def compute(real_data, synthetic_data, metadata, **kwargs):
        results = {}
        for rel in metadata.relationships:
            cardinality_real = get_cardinality_distribution(
                real_data[rel["parent_table_name"]][rel["parent_primary_key"]],
                real_data[rel["child_table_name"]][rel["child_foreign_key"]],
            )
            cardinality_synthetic = get_cardinality_distribution(
                synthetic_data[rel["parent_table_name"]][rel["parent_primary_key"]],
                synthetic_data[rel["child_table_name"]][rel["child_foreign_key"]],
            )
            statistic, pval = ks_2samp(cardinality_real, cardinality_synthetic)
            results[f"{rel['parent_table_name']}_{rel['child_table_name']}"] = {
                "statistic": statistic,
                "pval": pval,
            }
        return results



"""Dataset Generators to test the rdt2 Transformers."""

from collections import defaultdict

from rctgan.rdt2.performance.datasets import boolean, categorical, datetime, numerical, pii
from rctgan.rdt2.performance.datasets.base import BaseDatasetGenerator

__all__ = [
    'boolean',
    'categorical',
    'datetime',
    'numerical',
    'BaseDatasetGenerator',
    'pii',
]


def get_dataset_generators_by_type():
    """Build a ``dict`` mapping sdtypes to dataset generators.

    Returns:
        dict:
            Mapping of sdtype to a list of dataset generators that produce
            data of that sdtype.
    """
    dataset_generators = defaultdict(list)
    for dataset_generator in BaseDatasetGenerator.get_subclasses():
        dataset_generators[dataset_generator.SDTYPE].append(dataset_generator)

    return dataset_generators



"""Hierarchical Modeling Algorithms."""

import itertools
import logging
import pickle

import numpy as np
import pandas as pd

from rctgan.errors import NotFittedError
from rctgan.metadata import Metadata, utils

LOGGER = logging.getLogger(__name__)


class BaseRelationalModel:
    """Base class for all the relational models.

    The ``BaseRelationalModel`` class defines the common API that all the
    relational models need to implement, as well as common functionality.

    Args:
        metadata (dict, str or Metadata):
            Metadata dict, path to the metadata JSON file or Metadata instance itself.
        root_path (str or None):
            Path to the dataset directory. If ``None`` and metadata is
            a path, the metadata location is used. If ``None`` and
            metadata is a dict, the current working directory is used.
    """

    metadata = None

    def __init__(self, metadata, root_path=None):
        if isinstance(metadata, Metadata):
            self.metadata = metadata
        else:
            self.metadata = Metadata(metadata, root_path)

        self._primary_key_generators = dict()
        self._remaining_primary_keys = dict()

    def _fit(self, tables=None):
        """Fit this relational model instance to the dataset data.

        Args:
            tables (dict):
                Dictionary with the table names as key and ``pandas.DataFrame`` instances as
                values.  If ``None`` is given, the tables will be loaded from the paths
                indicated in ``metadata``. Defaults to ``None``.
        """
        raise NotImplementedError()

    def fit(self, tables=None):
        """Fit this relational model instance to the dataset data.

        Args:
            tables (dict):
                Dictionary with the table names as key and ``pandas.DataFrame`` instances as
                values.  If ``None`` is given, the tables will be loaded from the paths
                indicated in ``metadata``. Defaults to ``None``.
        """
        self._fit(tables)
        self.fitted = True

    def _reset_primary_keys_generators(self):
        """Reset the primary key generators."""
        self._primary_key_generators = dict()
        self._remaining_primary_keys = dict()

    def _get_primary_keys(self, table_name, num_rows):
        """Return the primary key and amount of values for the requested table.

        Args:
            table_name (str):
                Name of the table to get the primary keys from.
            num_rows (str):
                Number of ``primary_keys`` to generate.

        Returns:
            tuple (str, pandas.Series):
                primary key name and primary key values. If the table has no primary
                key, ``(None, None)`` is returned.

        Raises:
            ValueError:
                If the ``metadata`` contains invalid types or subtypes, or if
                there are not enough primary keys left on any of the generators.
            NotImplementedError:
                If the primary key subtype is a ``datetime``.
        """
        primary_key = self.metadata.get_primary_key(table_name)

        field = self.metadata.get_fields(table_name)[primary_key]

        generator = self._primary_key_generators.get(table_name)

        if generator is None:
            if field['type'] != 'id':
                raise ValueError('Only columns with type `id` can be primary keys')

            subtype = field.get('subtype', 'integer')
            if subtype == 'integer':
                generator = itertools.count()
                remaining = np.inf
            elif subtype == 'string':
                regex = field.get('regex', r'^[a-zA-Z]+$')
                generator, remaining = utils.strings_from_regex(regex)
            elif subtype == 'datetime':
                raise NotImplementedError('Datetime ids are not yet supported')
            else:
                raise ValueError('Only `integer` or `string` id columns are supported.')

            self._primary_key_generators[table_name] = generator
            self._remaining_primary_keys[table_name] = remaining

        else:
            remaining = self._remaining_primary_keys[table_name]

        if remaining < num_rows:
            raise ValueError(
                'Not enough unique values for primary key of table {}'
                ' to generate {} samples.'.format(table_name, num_rows)
            )

        self._remaining_primary_keys[table_name] -= num_rows
        primary_key_values = pd.Series([x for i, x in zip(range(num_rows), generator)])

        return primary_key_values

    def _sample(self, table_name=None, num_rows=None, sample_children=True):
        """Generate synthetic data for one table or the entire dataset."""
        raise NotImplementedError()

    def sample(self, table_name=None, num_rows=None,
               sample_children=True, reset_primary_keys=False):
        """Generate synthetic data for one table or the entire dataset.

        If a ``table_name`` is given and ``sample_children`` is ``False``, a
        ``pandas.DataFrame`` with the values from the indicated table is returned.
        Otherwise, if ``sample_children`` is ``True``, a dictionary containing both
        the table and all its descendant tables is returned.

        If no ``table_name`` is given, the entire dataset is sampled and returned
        in a dictionary.

        If ``num_rows`` is given, the root tables of the dataset will contain the
        indicated number of rows. Otherwise, the number of rows will be the same
        as in the original dataset. Number of rows in the child tables cannot be
        controlled and always will depend on the values from the sampled parent
        tables.

        If ``reset_primary_keys`` is ``True``, the primary key generators will be
        reset.

        Args:
            table_name (str):
                Name of the table to sample from. If not passed, sample the entire
                dataset.
            num_rows (int):
                Amount of rows to sample. If ``None``, sample the same number of rows
                as there were in the original table.
            sample_children (bool):
                Whether or not sample child tables. Used only if ``table_name`` is
                given. Defaults to ``True``.
            reset_primary_keys (bool):
                Whether or not reset the primary keys generators. Defaults to ``False``.

        Returns:
            dict or pandas.DataFrame:
                - Returns a ``dict`` when ``sample_children`` is ``True`` with the sampled table
                  and child tables.
                - Returns a ``pandas.DataFrame`` when ``sample_children`` is ``False``.

        Raises:
            NotFittedError:
                A ``NotFittedError`` is raised when the model has not been fitted yet.
        """
        if not self.fitted:
            raise NotFittedError('SDV instance has not been fitted')

        if reset_primary_keys:
            self._reset_primary_keys_generators()

        return self._sample(table_name, num_rows, sample_children)

    def save(self, path):
        """Save this instance to the given path using pickle.

        Args:
            path (str):
                Path where the instance will be serialized.
        """
        with open(path, 'wb') as output:
            pickle.dump(self, output)

    @classmethod
    def load(cls, path):
        """Load a model from a given path.

        Args:
            path (str):
                Path from which to load the instance.
        """
        with open(path, 'rb') as f:
            return pickle.load(f)



from copy import deepcopy

import pandas as pd

from syntherela.metadata import drop_ids
from syntherela.metrics.base import DetectionBaseMetric, SingleTableMetric
from .denormalized_detection import DenormalizedDetection
from .parent_child import ParentChildDetection


class BaseAggregationDetection(DetectionBaseMetric):
    @staticmethod
    def add_aggregations(data, metadata, update_metadata=True):
        aggregated_data = deepcopy(data)
        for relationship in metadata.relationships:
            parent_table_name = relationship["parent_table_name"]
            child_table_name = relationship["child_table_name"]
            parent_column = relationship["parent_primary_key"]
            child_fk = relationship["child_foreign_key"]

            # add child counts
            child_df = pd.DataFrame(
                {
                    f"{child_table_name}_{child_fk}_counts": data[child_table_name][
                        child_fk
                    ].value_counts()
                }
            )
            cardinality_df = (
                pd.DataFrame({"parent": data[parent_table_name][parent_column]})
                .join(child_df, on="parent")
                .fillna(0)
            )
            aggregated_data[parent_table_name] = (
                aggregated_data[parent_table_name]
                .merge(
                    cardinality_df, how="left", left_on=parent_column, right_on="parent"
                )
                .drop(columns="parent")
            )

            if update_metadata:
                metadata.add_column(
                    parent_table_name,
                    f"{child_table_name}_{child_fk}_counts",
                    sdtype="numerical",
                )

            # add categorical counts
            categorical_columns = []
            for column_name, column_info in (
                metadata.tables[child_table_name].to_dict()["columns"].items()
            ):
                if column_info["sdtype"] == "categorical":
                    categorical_columns.append(column_name)

            if len(categorical_columns) > 0:
                categorical_df = data[child_table_name][
                    categorical_columns + [child_fk]
                ]
                categorical_column_names = [
                    f"{child_table_name}_{child_fk}_{column}_nunique"
                    for column in categorical_columns
                ]
                categorical_df.columns = categorical_column_names + [child_fk]

                aggregated_data[parent_table_name] = aggregated_data[
                    parent_table_name
                ].merge(
                    categorical_df.groupby(child_fk).nunique(),
                    how="left",
                    left_on=parent_column,
                    right_index=True,
                    suffixes=("", "_nunique"),
                )
                aggregated_data[parent_table_name][categorical_column_names] = (
                    aggregated_data[
                        parent_table_name
                    ][categorical_column_names].fillna(0)
                )

                if update_metadata:
                    for column in categorical_column_names:
                        metadata.add_column(
                            parent_table_name, column, sdtype="numerical"
                        )

            # add numerical means
            numerical_columns = []
            for column_name, column_info in (
                metadata.tables[child_table_name].to_dict()["columns"].items()
            ):
                if column_info["sdtype"] == "numerical":
                    numerical_columns.append(column_name)

            if len(numerical_columns) > 0:
                numerical_df = data[child_table_name][numerical_columns + [child_fk]]
                numerical_column_names = [
                    f"{child_table_name}_{child_fk}_{column}_mean"
                    for column in numerical_columns
                ]
                numerical_df.columns = numerical_column_names + [child_fk]

                aggregated_data[parent_table_name] = aggregated_data[
                    parent_table_name
                ].merge(
                    numerical_df.groupby(child_fk).mean(),
                    how="left",
                    left_on=parent_column,
                    right_index=True,
                    suffixes=("", "_mean"),
                )

                if update_metadata:
                    for column in numerical_column_names:
                        metadata.add_column(
                            parent_table_name, column, sdtype="numerical"
                        )

        return aggregated_data, metadata


class AggregationDetection(
    BaseAggregationDetection, DetectionBaseMetric, SingleTableMetric
):
    @staticmethod
    def is_applicable(metadata, table):
        """
        Check if the table contains at least one column that is not an id.
        And if the table has a relationship with another table.
        """
        nonid = False
        table_metadata = metadata.tables[table].to_dict()
        for column_name in table_metadata["columns"].keys():
            if table_metadata["columns"][column_name]["sdtype"] != "id":
                nonid = True
                break
        has_children = len(metadata.get_children(table)) > 0
        return nonid and has_children

    def run(self, real_data, synthetic_data, metadata, target_table=None, **kwargs):
        real_data_with_aggregations, updated_metadata = self.add_aggregations(
            real_data, deepcopy(metadata)
        )
        synthetic_data_with_aggregations, _ = self.add_aggregations(
            synthetic_data, metadata, update_metadata=False
        )
        results = {}
        if target_table is not None:
            table_metadata = metadata.tables[target_table].to_dict()
            real_data_with_aggregations[target_table] = drop_ids(
                real_data_with_aggregations[target_table], table_metadata
            )
            synthetic_data_with_aggregations[target_table] = drop_ids(
                synthetic_data_with_aggregations[target_table], table_metadata
            )
            return super().run(
                real_data_with_aggregations[target_table],
                synthetic_data_with_aggregations[target_table],
                metadata=updated_metadata,
                **kwargs,
            )

        for table in metadata.get_tables():
            table_metadata = metadata.tables[table].to_dict()
            if not self.is_applicable(updated_metadata, table):
                continue
            real_data_with_aggregations[table] = drop_ids(
                real_data_with_aggregations[table], table_metadata
            )
            synthetic_data_with_aggregations[table] = drop_ids(
                synthetic_data_with_aggregations[table], table_metadata
            )

            results[table] = super().run(
                real_data_with_aggregations[table],
                synthetic_data_with_aggregations[table],
                metadata=updated_metadata,
                **kwargs,
            )
        return results


class DenormalizedAggregationDetection(DenormalizedDetection, BaseAggregationDetection):
    def prepare_data(self, real_data, synthetic_data, metadata):
        aggregated_real_data, updated_metadata = self.add_aggregations(
            real_data, deepcopy(metadata)
        )
        aggregated_synthetic_data, _ = self.add_aggregations(
            synthetic_data, metadata, update_metadata=False
        )
        return super().prepare_data(
            aggregated_real_data, aggregated_synthetic_data, updated_metadata
        )


class ParentChildAggregationDetection(ParentChildDetection, BaseAggregationDetection):
    def prepare_data(
        self,
        real_data,
        synthetic_data,
        metadata,
        parent_table,
        child_table,
        pair_metadata,
    ):
        aggregated_real_data, updated_metadata = self.add_aggregations(
            real_data, deepcopy(metadata)
        )
        aggregated_synthetic_data, _ = self.add_aggregations(
            synthetic_data, metadata, update_metadata=False
        )
        return super().prepare_data(
            aggregated_real_data,
            aggregated_synthetic_data,
            updated_metadata,
            parent_table,
            child_table,
            pair_metadata,
        )



# -*- coding: utf-8 -*-
"""
Created on Mon Jun 20 14:09:24 2022

@author: mohamedg
"""
from rctgan.rdt2 import HyperTransformer
from rctgan.rdt2.transformers.numerical import FloatFormatter, GaussianNormalizer
from rctgan.rdt2.transformers.categorical import FrequencyEncoder, OneHotEncoder
from rctgan.rdt2.transformers.datetime import OptimizedTimestampEncoder
import itertools
from rctgan.tabular import CTGAN, PC_CTGAN
import pandas as pd
import numpy as np
import random
from scipy.stats import truncnorm, kstest
import sys

class RCTGAN:
    def __init__(self, metadata=None, hyperparam=None, current_table=None,
                 ohe_for_parent=False, if_gaussian_ht=True, num_transformers='gaussian', seed=None):
        self.metadata = metadata
        self.transformers = {}
        self.size_tables = {}
        self.size_stats = {}
        self.models = {}
        self.hyperparam = hyperparam
        self.default_hyperparam()
        self.current_table = current_table
        self.ohe_for_parent = ohe_for_parent
        self.if_gaussian_ht = if_gaussian_ht
        if self.ohe_for_parent:
            self.if_gaussian_ht = False
        if self.if_gaussian_ht==False:
            self.num_transformers = num_transformers
        elif num_transformers=='gaussian':
            self.num_transformers = 'float'
        else:
            self.num_transformers = num_transformers
        self.seed = seed
    
    def default_hyperparam(self):
        default_hyp = {"embedding_dim": 128,
                       "generator_dim": (256, 256),
                       "discriminator_dim": (256, 256),
                       "generator_lr": 2e-4,
                       "generator_decay": 1e-6,
                       "discriminator_lr": 2e-4,
                       "discriminator_decay": 1e-6,
                       "batch_size": 500,
                       "discriminator_steps": 1,
                       "log_frequency": True,
                       "verbose": False,
                       "epochs": 1000,
                       "pac": 10,
                       "cuda": True,
                       "plot_loss": False,
                       "grand_parent": True,
                       "field_transformers": None,
                       "anonymize_fields": None,
                       "constraints": None,
                       "table_metadata": None,
                       "rounding": 'auto', 
                       'min_value': 'auto', 
                       'max_value': 'auto'
                      }
        if self.hyperparam==None:
            self.hyperparam = {}
            for table_name in list(self.metadata.get_tables()):
                self.hyperparam[table_name] = default_hyp
        else:
            for table_name in list(self.metadata.get_tables()):
                if not table_name in list(self.hyperparam.keys()):
                    self.hyperparam[table_name] = default_hyp
                else:
                    for hyp in list(default_hyp.keys()):
                        if not hyp in list(self.hyperparam[table_name].keys()):
                            self.hyperparam[table_name][hyp] = default_hyp[hyp]
    
    
    def set_hyperparam(self, hyper):
        self.hyperparam = hyper
        self.default_hyperparam()
    
    def get_hyperparam(self):
        return self.hyperparam
    
    def set_tab_hyperameter(self, table_name, tab_hyper):
        self.hyperparam[table_name] = tab_hyper
        self.default_hyperparam()
    
    def rdt2_transform(self, meta_fields, table, field_deleted=[]): # meta_fields = metadata.get_table_meta('store')['fields']
        ht = HyperTransformer()
        config_dict = {"sdtypes": {}, "transformers": {}}
        col_retained = []
        for field in meta_fields.keys():
            if field not in field_deleted:
                if meta_fields[field]['type'] == 'categorical':
                    col_retained.append(field)
                    config_dict["sdtypes"][field] =  'categorical'
                    if self.ohe_for_parent:
                        config_dict["transformers"][field] =  OneHotEncoder()
                    else:
                        config_dict["transformers"][field] =  FrequencyEncoder(add_noise=True)
                elif meta_fields[field]['type'] == 'numerical':
                    col_retained.append(field)
                    config_dict["sdtypes"][field] =  'numerical'
                    if self.num_transformers=='gaussian':
                        config_dict["transformers"][field] =  GaussianNormalizer()
                    elif self.num_transformers=='float':
                        config_dict["transformers"][field] =  FloatFormatter(missing_value_replacement='mean')
                    else:
                        config_dict["transformers"][field] =  GaussianNormalizer()
                    
                elif meta_fields[field]['type'] == 'datetime':
                    col_retained.append(field)
                    config_dict["sdtypes"][field] =  'datetime'
                    format_datetime = meta_fields[field]['format']
                    config_dict["transformers"][field] =  OptimizedTimestampEncoder(missing_value_replacement='mean', datetime_format=format_datetime)
        ht.set_config(config=config_dict)
        ht.fit(table[col_retained])
        return ht, col_retained
    
    def gaussian_ht(self, table_transormed):
        ht = HyperTransformer()
        config_dict = {"sdtypes": {}, "transformers": {}}
        for col in table_transormed.columns:
            config_dict["sdtypes"][col] =  'numerical'
            config_dict["transformers"][col] =  GaussianNormalizer()
        ht.set_config(config=config_dict)
        ht.fit(table_transormed)
        return ht
    
    def transform(self, table_name, table):
        col = self.transformers[table_name]["columns"]
        if self.if_gaussian_ht==False:
            return self.transformers[table_name]['hypertr'].transform(table[col])
        else:
            return  self.transformers[table_name]['gaussian_ht'].transform(self.transformers[table_name]['hypertr'].transform(table[col]))

    def keep_data_col(self, meta_fields):
        col_retained = []
        for field in meta_fields.keys():
            if meta_fields[field]['type'] == 'categorical':
                col_retained.append(field)
            elif meta_fields[field]['type'] == 'numerical':
                col_retained.append(field)
            elif meta_fields[field]['type'] == 'datetime':
                col_retained.append(field)
        return col_retained
    
    def parents_input_add(self, table_name, tables, parents_trandformed=None):
        
        if str(parents_trandformed)=="None":
            count = 0
        else:
            count = len(parents_trandformed.columns)
        parents_name = list(self.metadata.get_parents(table_name))
        for parent_name in parents_name:
            parent_prim_key = self.metadata.get_primary_key(parent_name)
            foreign_keys = list(self.metadata.get_foreign_keys(parent_name, table_name))
            bool_test = False
            
            ht = self.transformers[parent_name]["hypertr"]
            col = self.transformers[parent_name]["columns"]

            for foreign_key in foreign_keys:
                parent_trandformed = self.transform(parent_name, tables[parent_name])
                parent_trandformed.columns = ["var_"+str(i) for i in range(1, len(parent_trandformed.columns)+1)]
                if self.hyperparam[self.current_table]["grand_parent"]:
                    if parent_name in list(self.metadata.get_parents(self.current_table)):
                        parent_trandformed = self.parents_input_add(parent_name, tables, parent_trandformed)
                parent_trandformed.columns = ["var_"+str(count+i) for i in range(1, len(parent_trandformed.columns)+1)]
                temp_serie = pd.DataFrame(tables[table_name][foreign_key].copy())
                parent_trandformed[foreign_key] = list(tables[parent_name][parent_prim_key])
                parent_trandformed = temp_serie.merge(parent_trandformed, 
                                                              on=[foreign_key],
                                                              how='left', 
                                                              indicator=True)
                parent_trandformed = parent_trandformed.drop([foreign_key, '_merge'], axis=1)
                
                if str(parents_trandformed)=="None":
                    parents_trandformed = parent_trandformed.copy()
                else:
                    parents_trandformed = parents_trandformed.join(parent_trandformed)
                count = len(parents_trandformed.columns)
                del temp_serie
            del parent_trandformed
        return parents_trandformed
    
    def fit(self, tables):
        if self.seed is not None:
            random.seed(self.seed)
            np.random.seed(self.seed)
        for table_name in self.metadata.get_tables():
            children = list(self.metadata.get_children(table_name))
            self.size_tables[table_name] =len(tables[table_name])
            if len(children)>0:
                meta = self.metadata.get_table_meta(table_name)['fields']
                ht, col = self.rdt2_transform(meta, tables[table_name])
                self.transformers[table_name] = {"hypertr": ht, "columns": col}
                del ht
                del col
                if self.if_gaussian_ht:
                    ht = self.transformers[table_name]['hypertr']
                    col = self.transformers[table_name]['columns']
                    self.transformers[table_name]['gaussian_ht'] = self.gaussian_ht(ht.transform(tables[table_name][col]))

                self.size_stats[table_name] = {}
            else:
                meta = self.metadata.get_table_meta(table_name)['fields']
                self.transformers[table_name] = {"columns": self.keep_data_col(meta)}

        for table_name in self.metadata.get_tables():
            self.current_table = table_name
            if len(self.metadata.get_parents(table_name)) == 0:
                prim_key = self.metadata.get_primary_key(table_name)
                model = CTGAN(primary_key=prim_key, 
                              embedding_dim=self.hyperparam[table_name]["embedding_dim"], 
                              generator_dim=self.hyperparam[table_name]["generator_dim"], 
                              discriminator_dim=self.hyperparam[table_name]["discriminator_dim"],
                              generator_lr=self.hyperparam[table_name]["generator_lr"], 
                              generator_decay=self.hyperparam[table_name]["generator_decay"], 
                              discriminator_lr=self.hyperparam[table_name]["discriminator_lr"],
                              discriminator_decay=self.hyperparam[table_name]["discriminator_decay"], 
                              batch_size=self.hyperparam[table_name]["batch_size"], 
                              discriminator_steps=self.hyperparam[table_name]["discriminator_steps"],
                              log_frequency=self.hyperparam[table_name]["log_frequency"], 
                              verbose=self.hyperparam[table_name]["verbose"], 
                              epochs=self.hyperparam[table_name]["epochs"], 
                              pac=self.hyperparam[table_name]["pac"], 
                              cuda=self.hyperparam[table_name]["cuda"],
                              plot_loss=self.hyperparam[table_name]["plot_loss"],
                              seed=self.seed,
                              field_transformers=self.hyperparam[table_name]["field_transformers"],
                              anonymize_fields=self.hyperparam[table_name]["anonymize_fields"],
                              constraints=self.hyperparam[table_name]["constraints"],
                              table_metadata=self.hyperparam[table_name]["table_metadata"],
                              rounding=self.hyperparam[table_name]["rounding"],
                              min_value=self.hyperparam[table_name]["min_value"],
                              max_value=self.hyperparam[table_name]["max_value"])
                col_table = self.transformers[table_name]["columns"]
                children = list(self.metadata.get_children(table_name))
                temp_table = tables[table_name][[prim_key]+col_table].copy()
                for child_name in children:
                    '''
                    Create the number of occurrence columns for the children of table_name.
                    Those columns count the number of children of each row for table_name.
                    they take part of the modeling and  will be usefull during sampling step for
                    chosing the number of children to generate of each sampled row in table_name.
                    '''
                    foreign_keys = list(self.metadata.get_foreign_keys(table_name, child_name))
                    for foreign_key in foreign_keys:
                        temp_child = pd.DataFrame(tables[child_name].groupby([foreign_key])[foreign_key].count())
                        temp_child.columns = [child_name+"_"+foreign_key+"_nb_occ"]
                        temp_child[prim_key] = temp_child.index
                        temp_child.index = range(len(temp_child))
                        temp_table = temp_table.merge(temp_child, 
                                                      on=[prim_key], 
                                                      how='left', 
                                                      indicator=True)
                        temp_table = temp_table.drop(['_merge'], axis=1)
                        mask = temp_table[child_name+"_"+foreign_key+"_nb_occ"].isna()
                        temp_table.loc[mask, child_name+"_"+foreign_key+"_nb_occ"] = 0
                        temp_table[child_name+"_"+foreign_key+"_nb_occ"] = temp_table[child_name+"_"+foreign_key+"_nb_occ"].astype(int)
                        self.size_stats[table_name][child_name+"_"+foreign_key+"_nb_occ"] = {"min": np.min(temp_table[child_name+"_"+foreign_key+"_nb_occ"]),
                                                                                             "max": np.max(temp_table[child_name+"_"+foreign_key+"_nb_occ"]),
                                                                                             "mean": np.mean(temp_table[child_name+"_"+foreign_key+"_nb_occ"]),
                                                                                             "std": np.std(temp_table[child_name+"_"+foreign_key+"_nb_occ"])
                                                                                            }
                if self.hyperparam[table_name]["plot_loss"]==True:
                    print("plot of table: "+table_name)

                model.fit(temp_table)
                self.models[table_name] = model
                del temp_table
            
            else:
                prim_key = self.metadata.get_primary_key(table_name)
                parents_trandformed = self.parents_input_add(table_name, tables)
                model = PC_CTGAN(embedding_dim=self.hyperparam[table_name]["embedding_dim"], 
                                 generator_dim=self.hyperparam[table_name]["generator_dim"], 
                                 discriminator_dim=self.hyperparam[table_name]["discriminator_dim"],
                                 generator_lr=self.hyperparam[table_name]["generator_lr"], 
                                 generator_decay=self.hyperparam[table_name]["generator_decay"], 
                                 discriminator_lr=self.hyperparam[table_name]["discriminator_lr"],
                                 discriminator_decay=self.hyperparam[table_name]["discriminator_decay"], 
                                 batch_size=self.hyperparam[table_name]["batch_size"], 
                                 discriminator_steps=self.hyperparam[table_name]["discriminator_steps"],
                                 log_frequency=self.hyperparam[table_name]["log_frequency"], 
                                 verbose=self.hyperparam[table_name]["verbose"], 
                                 epochs=self.hyperparam[table_name]["epochs"], 
                                 pac=self.hyperparam[table_name]["pac"], 
                                 cuda=self.hyperparam[table_name]["cuda"],
                                 plot_loss=self.hyperparam[table_name]["plot_loss"],
                                 seed=self.seed,
                                 field_transformers=self.hyperparam[table_name]["field_transformers"],
                                anonymize_fields=self.hyperparam[table_name]["anonymize_fields"],
                                constraints=self.hyperparam[table_name]["constraints"],
                                table_metadata=self.hyperparam[table_name]["table_metadata"],
                                rounding=self.hyperparam[table_name]["rounding"],
                                min_value=self.hyperparam[table_name]["min_value"],
                                max_value=self.hyperparam[table_name]["max_value"])
                    
                col_table = self.transformers[table_name]["columns"]
                children = list(self.metadata.get_children(table_name))
                if len(children)>0:
                    '''
                    Create the number of occurrence columns for the children of table_name.
                    Those columns count the number of children of each row for table_name.
                    they take part of the modeling and  will be usefull during sampling step for
                    chosing the number of children to generate of each sampled row in table_name.
                    '''
                    temp_table = tables[table_name][[prim_key]+col_table].copy()
                    for child_name in children:
                        foreign_keys = list(self.metadata.get_foreign_keys(table_name, child_name))
                        for foreign_key in foreign_keys:
                            temp_child = pd.DataFrame(tables[child_name].groupby([foreign_key])[foreign_key].count())
                            temp_child.columns = [child_name+"_"+foreign_key+"_nb_occ"]
                            temp_child[prim_key] = temp_child.index
                            temp_child.index = range(len(temp_child))
                            temp_table = temp_table.merge(temp_child, 
                                                          on=[prim_key], 
                                                          how='left', 
                                                          indicator=True)
                            temp_table = temp_table.drop(['_merge'], axis=1)
                            mask = temp_table[child_name+"_"+foreign_key+"_nb_occ"].isna()
                            temp_table.loc[mask, child_name+"_"+foreign_key+"_nb_occ"] = 0
                            temp_table[child_name+"_"+foreign_key+"_nb_occ"] = temp_table[child_name+"_"+foreign_key+"_nb_occ"].astype(int)
                            self.size_stats[table_name][child_name+"_"+foreign_key+"_nb_occ"] = {"min": np.min(temp_table[child_name+"_"+foreign_key+"_nb_occ"]),
                                                                                             "max": np.max(temp_table[child_name+"_"+foreign_key+"_nb_occ"]),
                                                                                             "mean": np.mean(temp_table[child_name+"_"+foreign_key+"_nb_occ"]),
                                                                                             "std": np.std(temp_table[child_name+"_"+foreign_key+"_nb_occ"])
                                                                                            }
                    temp_table = temp_table.drop([prim_key], axis=1)
                else:
                    temp_table = tables[table_name][col_table]
                if self.hyperparam[table_name]["plot_loss"]==True:
                    print("plot of table: "+table_name)

                model.fit(temp_table, parents_trandformed)
                self.models[table_name] = model
                
    def generate_letter_id(self, size):
        liste = []
        # removed n as nan ids were being generated
        letters = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 
                   'j', 'k', 'l', 'm', 'o', 'p', 'q', 'r',
                   's', 't', 'u', 'v', 'w', 'x', 'y', 'z']
        is_first = True
        boolean =True
        while boolean:
            if is_first:
                combin_letters = letters
                liste += combin_letters
                is_first = False
            else:
                combin_letters_list = list(itertools.product(combin_letters, letters))
                combin_letters = []
                for c in combin_letters_list:
                    combin_letters += [c[0]+c[1]]
                liste += combin_letters
                
            if len(liste) > size:
                liste = liste[:size]
                boolean = False
        return liste
            
        
    
    def parent_child_sample_mini(self, child, sampled_data, table_transformed, f_key_frame):
        sampled_data[child] = self.models[child].sample(list(f_key_frame["_size_"]), table_transformed)

        sampled_data[child] = sampled_data[child].merge(f_key_frame, 
                                                          on=["Parent_index"],
                                                          how='left', 
                                                          indicator=True)
        sampled_data[child] = sampled_data[child].drop(["_size_", "Parent_index", "_merge"], axis=1)

        prim_key = self.metadata.get_primary_key(child)
        if prim_key:
            if self.metadata.get_table_meta(child)['fields'][prim_key]['subtype']=='string':
                sampled_data[child][prim_key] = self.generate_letter_id(len(sampled_data[child]))
            elif self.metadata.get_table_meta(child)['fields'][prim_key]['subtype']=='integer':
                sampled_data[child][prim_key] = range(1, len(sampled_data[child])+1)
        
    
    def granp_parent_transform_add(self, parent_name, table_transformed, f_key, f_key_frame, sampled_data, tables_transformed):
        grand_parents = list(self.metadata.get_parents(parent_name))
        for grand_parent in grand_parents:
            gp_foreign_keys = list(self.metadata.get_foreign_keys(grand_parent, parent_name))
            for gp_foreign_key in gp_foreign_keys:
                start_var = len(table_transformed.columns)
                temp_serie = pd.DataFrame(sampled_data[parent_name][gp_foreign_key])
                temp_serie.columns = [gp_foreign_key]
                temp_table = tables_transformed[grand_parent].copy()
                temp_table.columns = ["var_"+str(start_var+i+1) for i in range(len(temp_table.columns))] 
                
                grand_parent_prim_key = self.metadata.get_primary_key(grand_parent)
                temp_table[gp_foreign_key] = sampled_data[grand_parent][grand_parent_prim_key]
                
                temp_serie = temp_serie.merge(temp_table, on=[gp_foreign_key], how='left', indicator=True)
                temp_serie = temp_serie.drop([gp_foreign_key, "_merge"], axis=1)
                
                parent_prim_key = self.metadata.get_primary_key(parent_name)
                temp_serie[f_key] = sampled_data[parent_name][parent_prim_key]
                table_transformed[f_key] = f_key_frame[f_key]
                table_transformed = table_transformed.merge(temp_serie, on=[f_key], how='left', indicator=True)
                table_transformed = table_transformed.drop([f_key, "_merge"], axis=1)
        return table_transformed
                
    def dupli_rows(self, data, size_list):
        if len(data)==len(size_list):
            df = data.copy()
            df.index = range(len(df))
            df['index_prim_key'] = range(len(df))
            size_list_2 = []
            for k in range(len(size_list)):
                s = size_list[k]
                size_list_2 += [k]*s
            index_prim_key = pd.DataFrame(size_list_2, columns=['index_prim_key'])
            df = index_prim_key.merge(df, on=['index_prim_key'], how='left')
            df = df.drop(['index_prim_key'], axis=1)
            return df
        else:
            return None


    def parent_child_sample(self, child, sampled_data, tables_transformed):
        parents_name = list(self.metadata.get_parents(child))
        all_parents_sampled = True
        for parent_name in parents_name:
            if not parent_name in list(sampled_data.keys()):
                all_parents_sampled = False
            else:
                if self.hyperparam[child]["grand_parent"]:
                    grand_parents = list(self.metadata.get_parents(parent_name))
                    for grand_parent in grand_parents:
                        if not grand_parent in list(sampled_data.keys()):
                            all_parents_sampled = False
                    
            if all_parents_sampled == False:
                break
            
            if parent_name not in tables_transformed.keys():
                table_transformed = self.transform(parent_name, sampled_data[parent_name])
                tables_transformed[parent_name] = table_transformed.copy()
                del table_transformed
        
        if all_parents_sampled:
            enc_parent = parents_name[0]
            enc_foreign_key = list(self.metadata.get_foreign_keys(enc_parent, child))[0]
            enc_nb_occ_name = child+"_"+enc_foreign_key+"_nb_occ"
            
            table_transformed = tables_transformed[enc_parent].copy()
            table_transformed.columns = ["var_"+str(i+1) for i in range(len(table_transformed.columns))]
            
            prim_enc = self.metadata.get_primary_key(enc_parent)
            f_key_frame = pd.DataFrame(sampled_data[enc_parent][[prim_enc, enc_nb_occ_name]])
            f_key_frame.columns = [enc_foreign_key, "_size_"]
            f_key_frame["Parent_index"] = list(f_key_frame.index)
            
            if self.hyperparam[child]["grand_parent"]:
                table_transformed = self.granp_parent_transform_add(enc_parent, table_transformed, enc_foreign_key, f_key_frame, sampled_data, tables_transformed)
            
            if len(parents_name)==1 and len(list(self.metadata.get_foreign_keys(enc_parent, child)))==1:
                self.parent_child_sample_mini(child, sampled_data, table_transformed, f_key_frame)
            else:
                table_transformed = self.dupli_rows(table_transformed, list(f_key_frame["_size_"]))
                f_key_frame = self.dupli_rows(f_key_frame, list(f_key_frame["_size_"]))
                f_key_frame["_size_"] = [1 for _ in range(len(f_key_frame))]
                f_key_frame["Parent_index"] = list(f_key_frame.index)

                for parent_name in parents_name:
                    foreign_keys = list(self.metadata.get_foreign_keys(parent_name, child))
                    if parent_name==enc_parent:
                        foreign_keys.remove(enc_foreign_key)
                    for foreign_key in foreign_keys:
                        start_var = len(table_transformed.columns)
                        indexes_chosen = list(random.choices(tables_transformed[parent_name].index, k=len(table_transformed)))
                        random_parent_rows_transformed = tables_transformed[parent_name].loc[indexes_chosen].copy()
                        random_parent_rows_transformed.columns = ["var_"+str(start_var+i+1) for i in range(len(random_parent_rows_transformed.columns))]
                        random_parent_rows_transformed.index = range(len(random_parent_rows_transformed))
                        table_transformed = pd.concat([table_transformed, random_parent_rows_transformed], axis=1)
                        del random_parent_rows_transformed
                        
                        prim_key = self.metadata.get_primary_key(parent_name)
                        f_key_frame[foreign_key] = list(sampled_data[parent_name].loc[indexes_chosen][prim_key])
                        if self.hyperparam[child]["grand_parent"]:
                            table_transformed = self.granp_parent_transform_add(parent_name, table_transformed, foreign_key, f_key_frame, sampled_data, tables_transformed)
                self.parent_child_sample_mini(child, sampled_data, table_transformed, f_key_frame)
            
        children = list(self.metadata.get_children(child))
        if len(children)>0:
            table_transformed = self.transform(child, sampled_data[child])
            tables_transformed[child] = table_transformed.copy()
            del table_transformed
        for child_2 in children:
            self.parent_child_sample(child_2, sampled_data, tables_transformed)
            
        
    def sample(self):
        sampled_data = {}
        tables_transformed = {}
        if self.seed is not None:
            random.seed(self.seed)
            np.random.seed(self.seed)
        for table_name in self.metadata.get_tables():
            if len(self.metadata.get_parents(table_name)) == 0:
                prim_key = self.metadata.get_primary_key(table_name)
                n_table = self.size_tables[table_name]
                sampled_data[table_name] = self.models[table_name].sample(n_table)
                children = list(self.metadata.get_children(table_name))
                if len(children)>0:
                    table_transformed = self.transform(table_name, sampled_data[table_name])
                    tables_transformed[table_name] = table_transformed.copy()
                    del table_transformed
                for child in children:
                    self.parent_child_sample(child, sampled_data, tables_transformed)
        
        for tab_name in sampled_data.keys():
            for c in sampled_data[tab_name].columns:
                if c[-7:] == '_nb_occ':
                    sampled_data[tab_name] = sampled_data[tab_name].drop([c], axis=1)
        return sampled_data




"""Miscellaneous utility functions."""


def display_tables(tables, max_rows=10, datetime_fmt='%Y-%m-%d %H:%M:%S', row=True):
    """Display mutiple tables side by side on a Jupyter Notebook.

    Args:
        tables (dict[str, DataFrame]):
            ``dict`` containing table names and pandas DataFrames.
        max_rows (int):
            Max rows to show per table. Defaults to 10.
        datetime_fmt (str):
            Format with which to display datetime columns.
    """
    # Import here to avoid making IPython a hard dependency
    from IPython.core.display import HTML

    names = []
    data = []
    for name, table in tables.items():
        table = table.copy()
        for column in table.columns:
            column_data = table[column]
            if column_data.dtype.kind == 'M':
                table[column] = column_data.dt.strftime(datetime_fmt)

        names.append('<td style="text-align:left"><b>{}</b></td>'.format(name))
        data.append('<td>{}</td>'.format(table.head(max_rows).to_html(index=False)))

    if row:
        html = '<table><tr>{}</tr><tr>{}</tr></table>'.format(
            ''.join(names),
            ''.join(data),
        )
    else:
        rows = [
            '<tr>{}</tr><tr>{}</tr>'.format(name, table)
            for name, table in zip(names, data)
        ]
        html = '<table>{}</table>'.format(''.join(rows))

    return HTML(html)



import os
import json
import argparse
from datetime import datetime
from functools import partial

import numpy as np
import pandas as pd
from sklearn.naive_bayes import GaussianNB
from sklearn.svm import SVC, SVR
import xgboost as xgb
from scipy.stats import spearmanr, kendalltau, weightedtau
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.neighbors import KNeighborsRegressor, KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from syntherela.metadata import Metadata
from syntherela.data import load_tables, remove_sdv_columns
from syntherela.metrics.utility import MachineLearningEfficacyMetric
from syntherela.utils import NpEncoder


## DATA LOADING
def load_rossmann(method, run="1"):
    dataset_name = "rossmann_subsampled"

    metadata = Metadata().load_from_json(
        f"{PROJECT_PATH}/data/original/{dataset_name}/metadata.json"
    )

    tables = load_tables(f"{PROJECT_PATH}/data/original/{dataset_name}/", metadata)
    tables_synthetic = load_tables(
        f"{PROJECT_PATH}/data/synthetic/{dataset_name}/{method}/{run}/sample1", metadata
    )

    tables, metadata = remove_sdv_columns(tables, metadata)
    tables_synthetic, metadata = remove_sdv_columns(
        tables_synthetic, metadata, update_metadata=False
    )

    tables_test = load_tables(
        f'{PROJECT_PATH}/data/original/{dataset_name.split("_")[0]}/', metadata
    )
    tables_test, _ = remove_sdv_columns(tables_test, metadata, update_metadata=False)
    # split the test data
    min_date = datetime.strptime("2014-10-01", "%Y-%m-%d")
    max_date = datetime.strptime("2014-11-01", "%Y-%m-%d")
    tables_test["historical"] = tables_test["historical"][
        (tables_test["historical"]["Date"] >= min_date)
        & (tables_test["historical"]["Date"] < max_date)
    ]

    return tables, tables_synthetic, tables_test, metadata


def load_airbnb(method, run="1"):
    dataset_name = "airbnb-simplified_subsampled"

    metadata = Metadata().load_from_json(
        f"{PROJECT_PATH}/data/original/{dataset_name}/metadata.json"
    )

    tables = load_tables(f"{PROJECT_PATH}/data/original/{dataset_name}/", metadata)
    tables_synthetic = load_tables(
        f"{PROJECT_PATH}/data/synthetic/{dataset_name}/{method}/{run}/sample1", metadata
    )

    tables, metadata = remove_sdv_columns(tables, metadata)
    for table in tables:
        if "index" in tables_synthetic[table].columns:
            tables_synthetic[table].drop(columns=["index"], inplace=True)
    tables_synthetic, metadata = remove_sdv_columns(
        tables_synthetic, metadata, update_metadata=False
    )

    tables_test = load_tables(
        f'{PROJECT_PATH}/data/original/{dataset_name.split("_")[0]}/', metadata
    )
    tables_test, _ = remove_sdv_columns(tables_test, metadata, update_metadata=False)

    # select users with at most 50 sessions
    sessions_count = (
        tables_test["sessions"].reset_index().groupby("user_id").index.count()
    )
    eligable_users = sessions_count[sessions_count <= 50].index
    no_sessions = tables_test["users"][
        ~tables_test["users"]["id"].isin(tables_test["sessions"]["user_id"])
    ]["id"]
    eligable_users = eligable_users.union(no_sessions)
    # select users that are not in the synthetic data
    eligable_users = eligable_users[~eligable_users.isin(tables["users"]["id"])]

    selected_users = np.random.choice(eligable_users, 2000, replace=False)
    tables_test["users"] = tables_test["users"][
        tables_test["users"]["id"].isin(selected_users)
    ]
    tables_test["sessions"] = tables_test["sessions"][
        tables_test["sessions"]["user_id"].isin(selected_users)
    ]

    return tables, tables_synthetic, tables_test, metadata


def load_walmart(method, run="1"):
    dataset_name = "walmart_subsampled"

    metadata = Metadata().load_from_json(
        f"{PROJECT_PATH}/data/original/{dataset_name}/metadata.json"
    )

    tables = load_tables(f"{PROJECT_PATH}/data/original/{dataset_name}/", metadata)
    tables_synthetic = load_tables(
        f"{PROJECT_PATH}/data/synthetic/{dataset_name}/{method}/{run}/sample1", metadata
    )

    tables, metadata = remove_sdv_columns(tables, metadata)
    tables_synthetic, metadata = remove_sdv_columns(
        tables_synthetic, metadata, update_metadata=False
    )

    # split the test data
    tables_test = load_tables(
        f'{PROJECT_PATH}/data/original/{dataset_name.split("_")[0]}/', metadata
    )
    tables_test, _ = remove_sdv_columns(tables_test, metadata, update_metadata=False)

    min_date = datetime.strptime("2012-01-01", "%Y-%m-%d")
    max_date = datetime.strptime("2012-02-01", "%Y-%m-%d")
    tables_test["depts"] = tables_test["depts"][
        (tables_test["depts"]["Date"] >= min_date)
        & (tables_test["depts"]["Date"] < max_date)
    ]
    tables_test["features"] = tables_test["features"][
        (tables_test["features"]["Date"] >= min_date)
        & (tables_test["features"]["Date"] < max_date)
    ]

    return tables, tables_synthetic, tables_test, metadata


## DATA PREPARATION
def process_rossmann(tables, metadata):
    df = tables["historical"].merge(tables["store"], on="Store")
    numerical_columns = []
    for table in metadata.get_tables():
        table_metadata = metadata.get_table_meta(table)
        for column, column_info in table_metadata["columns"].items():
            if column_info["sdtype"] == "numerical":
                numerical_columns.append(column)
            elif column_info["sdtype"] == "id":
                if column in df.columns and column != "Store":
                    df.drop(columns=[column], inplace=True)

    # drop the StateHoliday column as it is constant causing problems with standardization
    # drop the DayOfWeek as we will be aggregating the data by month
    df.drop(columns=["StateHoliday", "DayOfWeek"], inplace=True)

    # set the categories for categorical variables (in case methods do not generate certain values)
    df["Open"] = pd.Categorical(df["Open"], categories=["0", "1"])
    df["Promo"] = pd.Categorical(df["Promo"], categories=["0", "1"])
    df["SchoolHoliday"] = pd.Categorical(df["SchoolHoliday"], categories=["0", "1"])
    df["StoreType"] = pd.Categorical(df["StoreType"], categories=["a", "b", "c", "d"])
    df["Assortment"] = pd.Categorical(df["Assortment"], categories=["a", "b", "c"])
    df["PromoInterval"] = pd.Categorical(
        df["PromoInterval"].astype("object").fillna("missing"),
        categories=[
            "missing",
            "Jan,Apr,Jul,Oct",
            "Mar,Jun,Sept,Dec",
            "Feb,May,Aug,Nov",
        ],
    )
    # one-hot encode the categorical columns
    df = pd.get_dummies(
        df,
        columns=[
            "Open",
            "Promo",
            "SchoolHoliday",
            "StoreType",
            "Assortment",
            "PromoInterval",
        ],
        drop_first=True,
    )
    # aggregate the data by store and month
    df["Month"] = df["Date"].dt.month
    df["Year"] = df["Date"].dt.year
    df = df.groupby(["Store", "Month", "Year"]).mean().reset_index(drop=True)
    # drop the dates due to subsampling
    df.drop(columns=["Date"], inplace=True)

    df[numerical_columns] = df[numerical_columns].fillna(0)
    y = df.pop("Customers")

    # remove missing / infinite y values
    mask = y.isna() | y.isin([np.inf, -np.inf])
    X = df[~mask]
    y = y[~mask]
    return X, y


def process_airbnb(tables, metadata, categories):
    df = tables["users"].copy()
    df.drop(columns=["date_first_booking"], inplace=True)

    for table in metadata.get_tables():
        table_metadata = metadata.get_table_meta(table)
        for column, column_info in table_metadata["columns"].items():
            if column_info["sdtype"] == "categorical":
                if column in df.columns:
                    df[column] = pd.Categorical(
                        df[column], categories=categories[column]
                    )

    # impute missing values
    df[["age"]] = df[["age"]].fillna(0)

    # aggregate the sessions data by user_id to obtain average session duration
    sessions_data = (
        tables["sessions"][["secs_elapsed", "user_id"]].groupby("user_id").mean()
    )
    # add the sessions count
    sessions_count = (
        tables["sessions"]["user_id"].value_counts().rename("sessions_count")
    )
    # join the sessions data and count of sessions
    sessions_data = sessions_data.join(sessions_count, on="user_id")

    # merge the sessions data with the users data and fill missing values with 0
    df = df.merge(sessions_data, left_on="id", right_index=True, how="left")
    df[sessions_data.columns] = df[sessions_data.columns].fillna(0)
    df.drop(columns=["id"], inplace=True)

    y = df.pop("country_destination")
    X = df.copy()

    # convert y to binary variable determining if the user booked a trip or not
    y = y != "NDF"

    return X, y


def process_walmart(tables, metadata):
    df = (
        tables["depts"]
        .merge(tables["stores"], on="Store")
        .merge(tables["features"], on=["Store", "Date"], suffixes=("", "_y"))
    )
    df.drop(df.filter(regex="_y$").columns, axis=1, inplace=True)
    df.drop(columns=["Dept"], inplace=True)

    # one-hot encode the categorical columns
    df["Type"] = pd.Categorical(df["Type"], categories=["A", "B", "C"])
    df = pd.get_dummies(df, columns=["Type"])
    # obtain average daily sales across all departments
    df = df.groupby(["Store", "Date"]).mean().reset_index(drop=True)

    y = df.pop("Weekly_Sales")

    # remove missing / infinite y values
    mask = y.isna() | y.isin([np.inf, -np.inf])
    X = df[~mask]
    y = y[~mask]
    return X, y


## UTILITY
def load_dataset(dataset_name, method, run="1"):
    if dataset_name == "rossmann":
        tables, tables_synthetic, tables_test, metadata = load_rossmann(method, run=run)
        feature_engineering_function = process_rossmann
        target = ("historical", "Customers", "regression")
    elif dataset_name == "airbnb":
        tables, tables_synthetic, tables_test, metadata = load_airbnb(method, run=run)
        # Store the categories from the real data to avoid mistakes in case methods do not generate certain values
        categories = {}
        for _, table in tables_test.items():
            for column in table.columns:
                if table[column].dtype.name == "category":
                    categories[column] = table[column].cat.categories.tolist()
        feature_engineering_function = partial(process_airbnb, categories=categories)
        # feature_engineering_function = None
        target = ("users", "country_destination", "classification")
    elif dataset_name == "walmart":
        tables, tables_synthetic, tables_test, metadata = load_walmart(method, run=run)
        feature_engineering_function = process_walmart
        target = ("depts", "Weekly_Sales", "regression")
    else:
        raise ValueError(f"Dataset {dataset_name} not supported.")

    return (
        tables,
        tables_synthetic,
        tables_test,
        metadata,
        feature_engineering_function,
        target,
    )


methods = [
    "SDV",
    "RCTGAN",
    "REALTABFORMER",
    "MOSTLYAI",
    "GRETEL_ACTGAN",
    "GRETEL_LSTM",
    "CLAVADDPM",
]

datasets = [
    "rossmann",
    "airbnb",
    "walmart",
]

classifiers = {
    "regression": {
        "xgboost": xgb.XGBRegressor,
        "linear": LinearRegression,
        "random_forest": RandomForestRegressor,
        "decision_tree": DecisionTreeRegressor,
        "knn": KNeighborsRegressor,
        "svr": SVR,
        "mlp": MLPRegressor,
    },
    "classification": {
        "xgboost": xgb.XGBClassifier,
        "linear": LogisticRegression,
        "random_forest": RandomForestClassifier,
        "decision_tree": DecisionTreeClassifier,
        "knn": KNeighborsClassifier,
        "svc": SVC,
        "gaussian_nb": GaussianNB,
        "mlp": MLPClassifier,
    },
}

cls_args = {
    "xgboost": {"random_state": None, "seed": None},
    "linear": {},
    "random_forest": {"random_state": None},
    "decision_tree": {"random_state": None},
    "knn": {},
    "svr": {},
    "svc": {"random_state": None, "probability": True},
    "gaussian_nb": {},
    "mlp": {"random_state": None},
}


feature_selection_models = {"regression": ["xgboost"], "classification": ["xgboost"]}

if __name__ == "__main__":
    from dotenv import load_dotenv

    load_dotenv()

    PROJECT_PATH = os.getenv("PROJECT_PATH")

    args = argparse.ArgumentParser()
    args.add_argument(
        "--dataset-name",
        type=str,
        default="rossmann",
        choices=datasets,
        help="Dataset name to run the experiment on.",
    )
    args.add_argument(
        "--methods",
        "-m",
        default=None,
        help="List of methods to evaluate.",
        action="append",
    )
    args.add_argument("--seed", type=int, default=0, help="Seed for reproducibility.")
    args.add_argument(
        "--bootstrap-repetitions",
        type=int,
        default=100,
        help="Number of bootstrap samples.",
    )
    args.add_argument("--run", type=str, default="1", help="Run number.")
    args = args.parse_args()

    dataset_name = args.dataset_name
    if args.methods is None:
        methods_to_run = methods
    else:
        methods_to_run = args.methods
    seed = args.seed
    m = args.bootstrap_repetitions
    run = args.run

    results = {}
    results[dataset_name] = {}
    real_scores = {}
    for method in methods_to_run:
        print(f"Method: {method}, Dataset: {dataset_name}")
        results[dataset_name][method] = {}
        (
            tables,
            tables_synthetic,
            tables_test,
            metadata,
            feature_engineering_function,
            target,
        ) = load_dataset(dataset_name, method, run=run)
        task = target[2]
        for classifier, classifier_cls in classifiers[task].items():
            real_scores.setdefault(classifier, {})
            real_scores[classifier].setdefault("score", None)
            real_scores[classifier].setdefault("se", None)
            real_scores[classifier].setdefault("feature_importance", None)
            score_real = real_scores[classifier]["score"]
            se_real = real_scores[classifier]["se"]
            feature_importance_real = real_scores[classifier]["feature_importance"]
            classifier_args_ = cls_args[classifier]
            if "random_state" in classifier_args_:
                classifier_args_["random_state"] = seed
            if "seed" in classifier_args_:
                classifier_args_["seed"] = seed

            ml_metric = MachineLearningEfficacyMetric(
                target=target,
                feature_engineering_function=feature_engineering_function,
                classifier_cls=classifier_cls,
                random_state=seed,
                classifier_args=classifier_args_,
            )

            result = ml_metric.run(
                tables,
                tables_synthetic,
                metadata,
                tables_test,
                m=m,
                feature_importance=classifier in feature_selection_models[task],
                score_real=score_real,
                se_real=se_real,
                feature_importance_real=feature_importance_real,
            )
            real_scores[classifier]["score"] = result["real_score"]
            real_scores[classifier]["se"] = result["real_score_se"]
            real_scores[classifier]["feature_importance"] = result[
                "full_feature_importance_real"
            ]
            print(
                f"Classifier: {classifier} real_score: {result['real_score'] :.3f} ± {result['real_score_se']:.3f}, synthetic_score: {result['synthetic_score']:.3f} ± {result['synthetic_score_se']:.3f}"
            )
            importances_syn = result.pop("importance_synthetic", None)
            results[dataset_name][method][classifier] = result
            if classifier in feature_selection_models[task]:
                feature_importances_spearman = []
                feature_importances_tau = []
                feature_importances_weighted = []
                feature_names = np.array(result["feature_names"])

                full_feature_importance_real = np.array(
                    result["full_feature_importance_real"]
                )
                full_feature_importance_synthetic = np.array(
                    result["full_feature_importance_synthetic"]
                )
                real_rank = np.argsort(full_feature_importance_real)
                synthetic_rank = np.argsort(full_feature_importance_synthetic)
                true_features_spearman_rank = spearmanr(
                    feature_names[real_rank], feature_names[synthetic_rank]
                ).statistic
                real_feature_order = feature_names[real_rank]
                synthetic_feature_order = feature_names[synthetic_rank]
                true_features_tau_rank = kendalltau(
                    feature_names[real_rank], feature_names[synthetic_rank]
                ).statistic
                indexed_real_rank = np.arange(len(real_rank))
                indexed_synthetic_rank = np.array(
                    [real_rank.tolist().index(feature) for feature in synthetic_rank]
                )
                true_features_weighted_rank = weightedtau(
                    indexed_real_rank, indexed_synthetic_rank, rank=None
                ).statistic

                for i in range(m):
                    importance_syn = importances_syn[i]
                    synthetic_rank = np.argsort(importance_syn)
                    features_spearman_rank = spearmanr(
                        feature_names[real_rank], feature_names[synthetic_rank]
                    ).statistic
                    features_tau_rank = kendalltau(
                        feature_names[real_rank], feature_names[synthetic_rank]
                    ).statistic

                    indexed_real_rank = np.arange(len(real_rank))
                    indexed_synthetic_rank_boot = np.array(
                        [
                            real_rank.tolist().index(feature)
                            for feature in synthetic_rank
                        ]
                    )

                    features_weighted_rank = weightedtau(
                        indexed_real_rank, indexed_synthetic_rank_boot, rank=None
                    ).statistic
                    feature_importances_spearman.append(features_spearman_rank)
                    feature_importances_tau.append(features_tau_rank)
                    feature_importances_weighted.append(features_weighted_rank)

        # rank the classifiers
        real_classifier_rank = list(
            dict(
                sorted(
                    results[dataset_name][method].items(),
                    key=lambda x: x[1]["real_score"],
                    reverse=True,
                )
            ).keys()
        )
        synthetic_classifier_rank = list(
            dict(
                sorted(
                    results[dataset_name][method].items(),
                    key=lambda x: x[1]["synthetic_score"],
                    reverse=True,
                )
            ).keys()
        )
        print(f"Real data classifier rank: {real_classifier_rank}")
        print(f"Synthetic data classifier rank: {synthetic_classifier_rank}")

        true_classifier_rank_spearman = spearmanr(
            real_classifier_rank, synthetic_classifier_rank
        ).statistic
        true_classifier_rank_kendall = kendalltau(
            real_classifier_rank, synthetic_classifier_rank
        ).statistic

        indexed_real_rank = np.array(
            [
                len(real_classifier_rank) - real_classifier_rank.index(classifier)
                for classifier in real_classifier_rank
            ]
        )
        indexed_synthetic_rank = np.array(
            [
                len(real_classifier_rank) - real_classifier_rank.index(classifier)
                for classifier in synthetic_classifier_rank
            ]
        )
        true_classifier_rank_weighted = weightedtau(
            indexed_real_rank, indexed_synthetic_rank, rank=None
        ).statistic

        classifier_rank_array_spearman = []
        classifier_rank_array_kendall = []
        classifier_rank_array_weighted = []

        for bootstrap_index in range(m):
            synthetic_classifier_rank_boot = list(
                dict(
                    sorted(
                        results[dataset_name][method].items(),
                        key=lambda x: x[1]["synthetic_score_array"][bootstrap_index],
                        reverse=True,
                    )
                ).keys()
            )

            classifier_rank_array_spearman.append(
                spearmanr(
                    real_classifier_rank, synthetic_classifier_rank_boot
                ).statistic
            )
            classifier_rank_array_kendall.append(
                kendalltau(
                    real_classifier_rank, synthetic_classifier_rank_boot
                ).statistic
            )

            indexed_real_rank_boot = np.array(
                [
                    len(real_classifier_rank) - real_classifier_rank.index(classifier)
                    for classifier in real_classifier_rank
                ]
            )
            indexed_synthetic_rank_boot = np.array(
                [
                    len(real_classifier_rank) - real_classifier_rank.index(classifier)
                    for classifier in synthetic_classifier_rank_boot
                ]
            )
            classifier_rank_array_weighted.append(
                weightedtau(
                    indexed_real_rank_boot, indexed_synthetic_rank_boot, rank=None
                ).statistic
            )

        spearman_rank = spearmanr(
            list(real_classifier_rank), list(synthetic_classifier_rank)
        )
        results[dataset_name][method]["classifier_rank"] = spearman_rank.statistic
        results[dataset_name][method]["feature_importance_spearman"] = (
            true_features_spearman_rank
        )
        results[dataset_name][method]["feature_importance_spearman_mean"] = np.mean(
            feature_importances_spearman
        )
        results[dataset_name][method]["feature_importance_spearman_se"] = np.std(
            feature_importances_spearman
        ) / np.sqrt(m)
        results[dataset_name][method]["feature_importance_tau"] = true_features_tau_rank
        results[dataset_name][method]["feature_importance_tau_mean"] = np.mean(
            feature_importances_tau
        )
        results[dataset_name][method]["feature_importance_tau_se"] = np.std(
            feature_importances_tau
        ) / np.sqrt(m)
        results[dataset_name][method]["feature_importance_weighted"] = (
            true_features_weighted_rank
        )
        results[dataset_name][method]["feature_importance_weighted_mean"] = np.mean(
            feature_importances_weighted
        )
        results[dataset_name][method]["feature_importance_weighted_se"] = np.std(
            feature_importances_weighted
        ) / np.sqrt(m)
        results[dataset_name][method]["spearman"] = true_classifier_rank_spearman
        results[dataset_name][method]["spearman_mean"] = np.mean(
            classifier_rank_array_spearman
        )
        results[dataset_name][method]["spearman_se"] = np.std(
            classifier_rank_array_spearman
        ) / np.sqrt(m)
        results[dataset_name][method]["kendall"] = true_classifier_rank_kendall
        results[dataset_name][method]["kendall_mean"] = np.mean(
            classifier_rank_array_kendall
        )
        results[dataset_name][method]["kendall_se"] = np.std(
            classifier_rank_array_kendall
        ) / np.sqrt(m)
        results[dataset_name][method]["weighted"] = true_classifier_rank_weighted
        results[dataset_name][method]["weighted_mean"] = np.mean(
            classifier_rank_array_weighted
        )
        results[dataset_name][method]["weighted_se"] = np.std(
            classifier_rank_array_weighted
        ) / np.sqrt(m)
        results[dataset_name][method]["real_feature_order"] = real_feature_order
        results[dataset_name][method]["synthetic_feature_order"] = (
            synthetic_feature_order
        )

        print()
        print(
            f"Boot spearman: {np.mean(classifier_rank_array_spearman):.3f}±{np.std(classifier_rank_array_spearman) / np.sqrt(m):.4f}"
        )
        print(
            f"Boot kendall: {np.mean(classifier_rank_array_kendall) :.3f}±{np.std(classifier_rank_array_kendall) / np.sqrt(m):.4f}"
        )
        print(
            f"Boot weighted: {np.mean(classifier_rank_array_weighted) :.3f}±{np.std(classifier_rank_array_weighted) / np.sqrt(m):.4f}"
        )

        print(
            f"Feature importance spearman: {np.mean(feature_importances_spearman) :.3f}±{np.std(feature_importances_spearman) / np.sqrt(m):.4f}"
        )
        print(
            f"Feature importance tau: {np.mean(feature_importances_tau) :.3f}±{np.std(feature_importances_tau) / np.sqrt(m):.4f}"
        )
        print(
            f"Feature importance weighted: {np.mean(feature_importances_weighted) :.3f}±{np.std(feature_importances_weighted) / np.sqrt(m):.4f}"
        )
        print()

        if len(methods_to_run) < len(methods):
            with open(
                f"{PROJECT_PATH}/results/mle_{dataset_name}_{run}_{seed}_{method}.json",
                "w",
            ) as f:
                json.dump(results, f, indent=4, cls=NpEncoder)

    with open(f"{PROJECT_PATH}/results/mle_{dataset_name}_{run}_{seed}.json", "w") as f:
        json.dump(results, f, indent=4, cls=NpEncoder)



import numpy as np
import pandas as pd
import xgboost as xgb
from tqdm import tqdm
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split
from sdmetrics.single_table.detection import LogisticDetection

from syntherela.metadata import Metadata
from syntherela.data import load_tables, remove_sdv_columns
from syntherela.metrics.single_table.detection.single_table_detection import (
    SingleTableDetection,
)

sns.set_theme()


def load_data(dataset_name, target_table):
    metadata = Metadata().load_from_json(f"data/original/{dataset_name}/metadata.json")
    tables = load_tables(f"data/original/{dataset_name}/", metadata)
    tables, metadata = remove_sdv_columns(tables, metadata)

    table_meta = metadata.get_table_meta(target_table, to_dict=True)

    return tables, table_meta, metadata


def symulate_generation(tables, target_table, seed=None):
    table = tables[target_table]
    table_perfect, table_original = train_test_split(
        table, test_size=0.5, random_state=seed
    )
    table_shuffled = table_perfect.copy()
    for column in table_shuffled.columns:
        table_shuffled[column] = table_shuffled[column].sample(frac=1).values
    return table_perfect, table_original, table_shuffled


def initialize_metrics(seed):
    xgb_cls = xgb.XGBClassifier
    xgb_args = {"seed": seed}
    lin_cls = LogisticRegression
    lin_args = {"random_state": seed}

    dd_xgb = SingleTableDetection(
        classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=seed
    )
    ld = SingleTableDetection(
        classifier_cls=lin_cls, classifier_args=lin_args, random_state=seed
    )
    return dd_xgb, ld


datasets = [
    "airbnb-simplified_subsampled",
    "Biodegradability_v1",
    "imdb_MovieLens_v1",
    "rossmann_subsampled",
    "walmart_subsampled",
]

dataset_names = {
    "airbnb-simplified_subsampled": "Airbnb\n(users)",
    "Biodegradability_v1": "Biodegradability\n(molecule)",
    "imdb_MovieLens_v1": "IMDB\n(movies)",
    "rossmann_subsampled": "Rossmann\n(store)",
    "walmart_subsampled": "Walmart\n(depts)",
}

target_tables = ["users", "molecule", "movies", "store", "stores"]

## Data Copying


def symulate_data_copying(tables, target_table, seed=None, frac_copied=1.0):
    table = tables[target_table]
    table_perfect, table_original = train_test_split(
        table, test_size=0.5, random_state=seed
    )
    table_copied = pd.concat(
        [
            table_original.sample(frac=frac_copied),
            table_perfect.sample(frac=1 - frac_copied),
        ]
    )
    return table_perfect, table_original, table_copied


seed = 0
results = {}
for dataset_name, target_table in zip(datasets, target_tables):
    tables, table_meta, _ = load_data(dataset_name, target_table)

    dd_xgb_perfect = []
    dd_xgb_copied = []
    ld_perfect = []
    ld_copied = []

    for i in tqdm(range(100), desc=dataset_name):
        # "generate" data
        table_perfect, table_original, table_copied = symulate_data_copying(
            tables, target_table, seed=seed + i
        )
        # prepare metrics
        dd_xgb, _ = initialize_metrics(seed + i)
        ld = LogisticDetection()

        # DD(XGB)
        results_dd_xgb_copied = dd_xgb.run(
            table_original,
            table_copied,
            table_meta,
        )

        # LD
        results_ld_copied = ld.compute(
            table_original,
            table_copied,
            table_meta,
        )

        dd_xgb_copied.append(results_dd_xgb_copied["accuracy"])
        ld_copied.append((results_ld_copied - 2) / -2)

    results[dataset_name] = {
        "dd_xgb_perfect": dd_xgb_perfect,
        "dd_xgb_copied": dd_xgb_copied,
        "ld_perfect": ld_perfect,
        "ld_copied": ld_copied,
    }

copying_behaviour = {}
for dataset_name, target_table in zip(datasets, target_tables):
    tables, table_meta, _ = load_data(dataset_name, target_table)

    accuracies = []
    std = []
    fractions = []
    step = -0.1
    for f in tqdm(np.arange(1, step, step), desc=dataset_name):
        f = round(max(0.0, f), 2)
        accs = []
        for i in range(100):
            _, table_original, table_copied = symulate_data_copying(
                tables, target_table, seed=seed + i, frac_copied=f
            )
            results_dd_xgb_copied = dd_xgb.run(
                table_original,
                table_copied,
                table_meta,
            )
            accs.append(results_dd_xgb_copied["accuracy"])
        accuracies.append(np.mean(accs))
        std.append(np.std(accs))
        fractions.append(f)
    copying_behaviour[dataset_name] = {
        "accuracies": np.array(accuracies),
        "std": np.array(std),
        "fractions": np.array(fractions),
    }

colormap = plt.cm.tab20
fig, axes = plt.subplots(1, 2, figsize=(10, 6))


def plot(ax, results, i, color, width=0.48, label=""):
    mean = np.mean(results)
    se = np.std(results) / np.sqrt(len(results))
    ax.bar(i, mean, label=label, width=width, color=color)
    ax.errorbar(i, mean, yerr=se, fmt="", color="black")


for i, dataset in enumerate(datasets):
    if i == 0:
        label1 = "DD(XGB)"
        label2 = "LD - max(AUC, 1/2)"
    else:
        label1 = ""
        label2 = ""

    plot(axes[0], results[dataset]["ld_copied"], 0 + i * 1.5, colormap(1), label=label2)
    plot(
        axes[0],
        results[dataset]["dd_xgb_copied"],
        0.5 + i * 1.5,
        colormap(3),
        label=label1,
    )


axes[0].set_xticks([0.25, 1.75, 3.25, 4.75, 6.25])
axes[0].set_xticklabels([dataset_names[dataset] for dataset in datasets], rotation=45)
axes[0].hlines(0.5, -0.25, 6.8, color="red", linestyle="--")
axes[0].legend(loc="upper right", bbox_to_anchor=(0.7, 1))
axes[0].set_ylim(0, 1)


axes[0].set_ylabel("Classification Accuracy")
axes[0].set_title("Copied Data")

for dataset_name in datasets:
    accuracies = copying_behaviour[dataset_name]["accuracies"]
    std = np.array(copying_behaviour[dataset_name]["std"])
    fractions = copying_behaviour[dataset_name]["fractions"]
    axes[1].errorbar(
        fractions,
        accuracies,
        yerr=std / np.sqrt(10),
        fmt="o",
        label=dataset_names[dataset_name].split("\n")[0],
    )

axes[1].hlines(0.5, 0, 1, color="red", linestyle="--")
axes[1].set_xticks(np.arange(0, 1.1, 0.1))
axes[1].set_title("Copying Behaviour by Dataset")
axes[1].set_xlabel("Fraction of Copied Data")
axes[1].set_ylabel("Classification Accuracy")
axes[1].legend()

fig.tight_layout()
plt.savefig("results/figures/figure7.png", dpi=300)
plt.show()



import os
import sys
import pickle
import logging
import argparse
from pathlib import Path


import pandas as pd
from rctgan import Metadata
from rctgan.relational import RCTGAN


MODEL_NAME = "RCTGAN"

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="airbnb-simplified_subsampled")
args.add_argument("--real-data-path", type=str, default="data/original")
args.add_argument("--synthetic-data-path", type=str, default="data/synthetic")
args.add_argument("--model-save-path", type=str, default="checkpoints")
args.add_argument("--run-id", type=str, default="1")
args = args.parse_args()

dataset_name = args.dataset_name
real_data_path = args.real_data_path
synthetic_data_path = args.synthetic_data_path
model_save_path = args.model_save_path
run_id = args.run_id

logger = logging.getLogger(f"{MODEL_NAME}_logger")

logger.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.info(f"START LOGGING RUN {run_id}...")


logger.info(f"START {MODEL_NAME}...")


# utils
def load_tables(data_path, metadata):
    tables = {}
    for file_name in os.listdir(data_path):
        if not file_name.endswith(".csv"):
            continue
        table_name = file_name.split(".")[0]
        dtypes = {}
        parse_dates = []
        for column, column_info in metadata.to_dict()["tables"][table_name][
            "fields"
        ].items():
            if column_info["type"] == "categorical":
                dtypes[column] = "object"
            elif column_info["type"] == "boolean":
                dtypes[column] = "bool"
            elif column_info["type"] == "datetime":
                parse_dates.append(column)
            # for ids and numerical values let pandas infer the type
        table = pd.read_csv(
            f"{data_path}/{file_name}",
            low_memory=False,
            dtype=dtypes,
            parse_dates=parse_dates,
        )
        tables[table_name] = table
    return tables


def remove_sdv_columns(tables, metadata, update_metadata=True):
    """
    "_v1" Versions of the relational demo datasets in SDV have some columns that are not present in the original datasets.
    We created this function to remove these columns from the tables and the metadata.
    """
    for table_name, table in tables.items():
        for column in table.columns:
            if any(
                prefix in column
                for prefix in ["add_numerical", "nb_rows_in", "min(", "max(", "sum("]
            ):
                table = table.drop(columns=column, axis=1)

                if not update_metadata:
                    continue
                metadata._metadata["tables"][table_name]["fields"].pop(column)

        tables[table_name] = table
    metadata.validate(tables)
    return tables, metadata


def save_tables(tables, path):
    if not os.path.exists(path):
        os.makedirs(path)
    for table_name, table in tables.items():
        table.to_csv(os.path.join(path, f"{table_name}.csv"), index=False)


# GENERATE SYNTHETIC DATA ---------------------------------
logger.debug(f"Loading real data... for {dataset_name}")
metadata = Metadata(
    metadata=str(Path(real_data_path) / f"{dataset_name}/metadata_v0.json")
)
real_data = load_tables(Path(real_data_path) / f"{dataset_name}", metadata)
real_data, metadata = remove_sdv_columns(real_data, metadata)
logger.debug("Real data loaded")

synthetic_data = {}

model = RCTGAN(metadata=metadata)
logger.info("Fitting model...")
model.seed = int(run_id)
model.fit(real_data)

# SAVE MODEL CHECKPOINT -----------------------------------
logger.debug("Saving model checkpoint...")
model_path = os.path.join(model_save_path, MODEL_NAME, dataset_name)
os.makedirs(model_path, exist_ok=True)
model_path = os.path.join(model_path, f"model_{run_id}.pkl")
pickle.dump(model, open(model_path, "wb"))

# SAMPLE AND SAVE DATA ------------------------------------
logger.info("Sampling and saving synthetic data...")
for i in range(1, 4):
    model.seed = i
    synthetic_data = model.sample()
    save_data_path = (
        Path(synthetic_data_path) / dataset_name / MODEL_NAME / run_id / f"sample{i}"
    )
    save_tables(synthetic_data, save_data_path)


logger.info("COMPLETE GENERATION DONE.")



import numpy as np
import xgboost as xgb
from tqdm import tqdm
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split

from syntherela.metadata import Metadata
from syntherela.data import load_tables, remove_sdv_columns
from syntherela.metrics.single_table.detection.single_table_detection import (
    SingleTableDetection,
)

sns.set_theme()


def load_data(dataset_name, target_table):
    metadata = Metadata().load_from_json(f"data/original/{dataset_name}/metadata.json")
    tables = load_tables(f"data/original/{dataset_name}/", metadata)
    tables, metadata = remove_sdv_columns(tables, metadata)

    table_meta = metadata.get_table_meta(target_table, to_dict=True)

    return tables, table_meta, metadata


def symulate_generation(tables, target_table, seed=None):
    table = tables[target_table]
    table_perfect, table_original = train_test_split(
        table, test_size=0.5, random_state=seed
    )
    table_shuffled = table_perfect.copy()
    for column in table_shuffled.columns:
        table_shuffled[column] = table_shuffled[column].sample(frac=1).values
    return table_perfect, table_original, table_shuffled


def initialize_metrics(seed):
    xgb_cls = xgb.XGBClassifier
    xgb_args = {"seed": seed}
    lin_cls = LogisticRegression
    lin_args = {"random_state": seed}

    dd_xgb = SingleTableDetection(
        classifier_cls=xgb_cls, classifier_args=xgb_args, random_state=seed
    )
    ld = SingleTableDetection(
        classifier_cls=lin_cls, classifier_args=lin_args, random_state=seed
    )
    return dd_xgb, ld


datasets = [
    "airbnb-simplified_subsampled",
    "Biodegradability_v1",
    "imdb_MovieLens_v1",
    "rossmann_subsampled",
    "walmart_subsampled",
]

dataset_names = {
    "airbnb-simplified_subsampled": "Airbnb\n(users)",
    "Biodegradability_v1": "Biodegradability\n(molecule)",
    "imdb_MovieLens_v1": "IMDB\n(movies)",
    "rossmann_subsampled": "Rossmann\n(store)",
    "walmart_subsampled": "Walmart\n(depts)",
}

target_tables = ["users", "molecule", "movies", "store", "stores"]
## Shuffling data

seed = 0
results = {}
for dataset_name, target_table in zip(datasets, target_tables):
    tables, table_meta, _ = load_data(dataset_name, target_table)

    dd_xgb_perfect = []
    dd_xgb_shuffled = []
    ld_perfect = []
    ld_shuffled = []

    for i in tqdm(range(100), desc=dataset_name):
        # "generate" data
        table_perfect, table_original, table_shuffled = symulate_generation(
            tables, target_table, seed=seed + i
        )
        # prepare metrics
        dd_xgb, ld = initialize_metrics(seed + i)

        # DD(XGB)
        results_dd_xgb_perfect = dd_xgb.run(
            table_original,
            table_perfect,
            table_meta,
        )

        results_dd_xgb_shuffled = dd_xgb.run(
            table_original,
            table_shuffled,
            table_meta,
        )

        # LD
        results_ld_perfect = ld.run(
            table_original,
            table_perfect,
            table_meta,
        )

        results_ld_shuffled = ld.run(
            table_original,
            table_shuffled,
            table_meta,
        )

        dd_xgb_perfect.append(results_dd_xgb_perfect["accuracy"])
        dd_xgb_shuffled.append(results_dd_xgb_shuffled["accuracy"])
        ld_perfect.append(results_ld_perfect["accuracy"])
        ld_shuffled.append(results_ld_shuffled["accuracy"])

    results[dataset_name] = {
        "dd_xgb_perfect": dd_xgb_perfect,
        "dd_xgb_shuffled": dd_xgb_shuffled,
        "ld_perfect": ld_perfect,
        "ld_shuffled": ld_shuffled,
    }


colormap = plt.cm.tab20
fig, axes = plt.subplots(1, 2, figsize=(10, 6))


def plot(ax, results, i, color, width=0.48, label=""):
    mean = np.mean(results)
    se = np.std(results) / np.sqrt(len(results))
    ax.bar(i, mean, label=label, width=width, color=color)
    ax.errorbar(i, mean, yerr=se, fmt="", color="black")


for i, dataset in enumerate(datasets):
    if i == 0:
        label1 = "DD(XGB)"
        label2 = "LD"
    else:
        label1 = ""
        label2 = ""
    plot(
        axes[0], results[dataset]["ld_perfect"], 0 + i * 1.5, colormap(1), label=label2
    )
    plot(
        axes[0],
        results[dataset]["dd_xgb_perfect"],
        0.5 + i * 1.5,
        colormap(3),
        label=label1,
    )

    plot(
        axes[1], results[dataset]["ld_shuffled"], 0 + i * 1.5, colormap(1), label=label2
    )
    plot(
        axes[1],
        results[dataset]["dd_xgb_shuffled"],
        0.5 + i * 1.5,
        colormap(3),
        label=label1,
    )


axes[0].set_xticks([0.25, 1.75, 3.25, 4.75, 6.25])
axes[0].set_xticklabels([dataset_names[dataset] for dataset in datasets], rotation=45)

axes[1].set_xticks([0.25, 1.75, 3.25, 4.75, 6.25])
axes[1].set_xticklabels([dataset_names[dataset] for dataset in datasets], rotation=45)

for ax in axes:
    ax.legend(loc="upper right", bbox_to_anchor=(0.7, 1))
    ax.set_ylim(0.3, 1)
    ax.hlines(0.5, -0.25, 6.8, color="red", linestyle="--")

axes[0].set_ylabel("Classification Accuracy")
axes[0].set_title("Perfectly Generated Data")
axes[1].set_title("Shuffled Data")

fig.tight_layout()
plt.savefig("results/figures/figure6.png", dpi=300)



import os
import argparse

import tqdm
import pandas as pd
from realtabformer import REaLTabFormer
import logging
import glob
import wandb
from relsyndgb.metadata import Metadata

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="rossmann")
args.add_argument("--full_sensitivity", type=bool, default=True)
args.add_argument("--retrain", type=bool, default=True)
args = args.parse_args()
dataset_name = args.dataset_name
full_sensitivity = args.full_sensitivity
retrain = args.retrain

logging.basicConfig(
    level=logging.DEBUG,  # Set the desired log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
    format='%(asctime)s - %(levelname)s - %(message)s',
    filename=f"REALTABFORMER_{dataset_name}.log",  # Specify the file where logs should be stored
)

root_table_name = sdv_metadata.get_root_table(dataset_name)

os.environ["WANDB_PROJECT"]=f"REALTABFORMER_{dataset_name}"

# save your trained model checkpoint to wandb
os.environ["WANDB_LOG_MODEL"]="true"

# turn off watch to log faster
os.environ["WANDB_WATCH"]="false"


logging.debug("START REALTABFORMER...")
logging.info("START INFO...")

batch_size_parent = 1024
batch_size_child = 10

# GENERATE SYNTHETIC DATA
logging.error("Generating synthetic data for %s, fold %d", dataset_name, k)
synthetic_data = {}

tables_train, tables_test = utils.get_train_test_split(dataset_name, k)
metadata = sdv_metadata.generate_metadata(dataset_name, tables_train)

# GENERATE PARENT TABLE
parent_df = tables_train[root_table_name]
logging.error(parent_df.head())
# check if join key is in both tables
join_on = metadata.to_dict()['tables'][root_table_name]['primary_key']

parent_model_path = f'models/realtabformer/{dataset_name}/checkpoint_{root_table_name}_{k}' 
# init parent model
if retrain or not(os.path.exists(parent_model_path) and len(os.listdir(parent_model_path)) > 0):
    parent_model = REaLTabFormer(model_type="tabular", batch_size=batch_size_parent, report_to="wandb")
    os.makedirs(parent_model_path, exist_ok=True)
    # fit and save parent model
    parent_model.fit(parent_df.drop(join_on, axis=1), full_sensitivity=full_sensitivity)
    parent_model.save(parent_model_path)
# load trained parent model
directories = list(filter(os.path.isdir, glob.glob(f"{parent_model_path}/id*")))
directories.sort(key=lambda x: os.path.getmtime(x))
parent_model_path = directories[-1]
parent_model = REaLTabFormer(
    model_type="tabular",
    parent_realtabformer_path=parent_model_path,
)
parent_model = parent_model.load_from_dir(parent_model_path)
# sample from parent model
parent_samples = parent_model.sample(tables_test[root_table_name].shape[0])
parent_samples.index.name = join_on
parent_samples = parent_samples.reset_index()
synthetic_data[root_table_name] = parent_samples

#GENERATE CHILD TABLES
for child_table_name in metadata.get_children(root_table_name):
    child_model_path = f'models/realtabformer/{dataset_name}/checkpoint_{child_table_name}_{k}'
    child_df = tables_train[child_table_name]
    logging.error(child_df.head())
    assert ((join_on in parent_df.columns) and
            (join_on in child_df.columns))
    # init child model
    if retrain or not(os.path.exists(child_model_path) and len(os.listdir(child_model_path)) > 0):
        child_model = REaLTabFormer(
            model_type="relational",
            parent_realtabformer_path=parent_model_path,
            output_max_length=None,
            train_size=0.95,
            batch_size=batch_size_child,
            report_to="wandb")
        # fit child model
        child_model.fit(
            df = child_df,
            in_df=parent_df,
            join_on=join_on)
        # save child model
        os.makedirs(child_model_path, exist_ok=True)
        child_model.save(child_model_path)

    directories = list(filter(os.path.isdir, glob.glob(f"{child_model_path}/id*")))
    directories.sort(key=lambda x: os.path.getmtime(x))
    child_model_path = directories[-1]
    child_model = REaLTabFormer(
        model_type="relational",
            parent_realtabformer_path=parent_model_path,
            output_max_length=None,
            train_size=0.95,
            batch_size=batch_size_child,
            report_to="wandb"
    )
    child_model = child_model.load_from_dir(child_model_path)
    
    child_samples = child_model.sample(
        input_unique_ids=parent_samples[join_on],
        input_df=parent_samples.drop(join_on, axis=1),
        gen_batch=64)
    
    child_samples.index.name = join_on
    child_samples = child_samples.reset_index()
    
    synthetic_data[child_table_name] = child_samples
    

utils.save_data(dataset_name=dataset_name, tables_synthetic=synthetic_data, leave_out_fold_num=k, method='REaLTabFormer', index=False)
